#copyright ReportLab Inc. 2001
#see license.txt for license details
#history www.reportlab.co.uk/rl-cgi/viewcvs.cgi/rlextra/pageCatcher/pageCatcher.py
"capture pages from PDF files, SEE script_usage FOR USAGE."

## EXAMPLE RUN (FEB 2001)
##C:\reportlab\repository\rlextra\pageCatcher>pageCatcher.py MakeForms testfile.pdf dump.data prefix --test testoutput.pdf 1 2
##Makeforms from testfile.pdf to dump.data page numbers [1, 2]
##
##
## placing pages [1, 2] from testfile.pdf into dump.data
##storing to dump.data
##making test grid from dump.data into testoutput.pdf
##now unpickling
##checking license
##license check failed
##form prefix1
##form prefix2
##wrote test file testoutput.pdf
##
##created forms ['prefix1', 'prefix2'] in dump.data
##
##C:\reportlab\repository\rlextra\pageCatcher>

import os

class UnresolvedRefError(ValueError):
    pass

# for convenience in the demo modes the user password is passed as a global variable
#  (in api usage it must be passed as a parameter)
PASSWORD = ""

from reportlab.rl_config import pageCompression, verbose, invariant, _FUZZ

DECOMPRESSSTREAMS=1 #set to auto decompress on input (recommended!)
DECOMPRESSIMAGES=0
COMPACT = pageCompression # set to compress on output

import sys, string, pickle, traceback

from pprint import pprint

from reportlab.pdfbase import pdfdoc, pdfutils
from reportlab.pdfgen import canvas
from reportlab.lib.utils import getStringIO, getBytesIO, open_and_read, _digester, asNative, asBytes,\
                md5, rl_reraise, int2Byte, rl_exec, isPy3, isNative, isNonPrimitiveInstance, strTypes
from reportlab.lib.boxstuff import aspectRatioFix
from reportlab.graphics.shapes import translate, rotate, zTransformPoint, inverse, transformPoint, mmult

if isPy3:
    def nativeFromBytes(x):
        if isinstance(x,bytes):
            return x.decode('latin1')
        elif isinstance(x,str):
            return x
        raise ValueError('nativeFromBytes(%s) called' % type(x))
    def bytesFromNative(x):
        if isinstance(x,bytes):
            return x
        elif isinstance(x,str):
            return x.encode('latin1')
        raise ValueError('bytesFromNative(%s) called' % type(x))
else:
    def nativeFromBytes(x):
        if isinstance(x,str):
            return x
        elif isinstance(x,unicode):
            return x.encode('latin1')
        raise ValueError('nativeFromBytes(%s) called' % type(x))
    def bytesFromNative(x):
        if isinstance(x,str):
            return x
        elif isinstance(x,unicode):
            return x.encode('latin1')
        raise ValueError('bytesFromNative(%s) called' % type(x))

# hack to support both kinds of x objects we have defined
if not hasattr(pdfdoc,'xObjectName'): pdfdoc.xObjectName = pdfdoc.formName

CHECK = {}

RERAISE = 1 # SET THIS TO GET A FULL TRACEBACK FROM main()

PAGESNAME = pdfdoc.PDFName("Pages")
PAGENAME = pdfdoc.PDFName("Page")

ESCAPEDICT = {
    "\\" : "\\",
    "(" : "(",
    ")" : ")",
    "r" : "\r",
    "n" : "\n",
    "t" : "\t",
    "" : "\b",
    "f" : "\f",
    }
OCTALDIGITS = "01234567"

class NULL:
    pass

def unescapePDFText(t):
    # first nuke all whitespace
    t = ''.join(t.strip().split())
    if t[0]!='<' and t[-1]!='>':
        raise ValueError("string must be of form '<..>'")
    outL = [].append
    n = len(t)
    if len(t)%2!=0:
        #print t
        raise ValueError("string content must have even number of digits")
    for i in range(1,n-1,2):
        outL(int2Byte(int(t[i:i+2], 16)))
    return ''.join(outL.__self__)

def unescapePDFString(s):
    #print; print "unescaping", s; stop
    if "\\" not in s: # shortcut, no tranlate needed
        if s[0]!="(" or s[-1]!=')':
            raise ValueError("all PDF strings must be bounded by parens (..)")
        return s[1:-1]
    inlist = list(s)
    outlist = []
    if inlist[0]!='(' or inlist[-1]!=')':
        raise ValueError("all PDF strings must be bounded by parens (..)")
    cursor = 1
    last = len(inlist)-1
    while cursor<last:
        thischar = inlist[cursor]
        #print "unescape at", ascii(thischar)
        cursor += 1
        if thischar=="\\":
            #print "escape!"
            next = inlist[cursor] # must have a next, otherwise not well formed
            cursor += 1
            if next in ESCAPEDICT:
                #print "escapedict", ascii(next)
                #print cursor,len(outlist), "alpha escape", thischar+next, "-->", ascii(ESCAPEDICT[next])
                thischar = ESCAPEDICT[next]
                #cursor += 1
            elif next in "\r\n": # skip end of line
                thischar = "" # no char
                # must have next, else not well formed
                next = inlist[cursor]
                if next in "\r\n":
                    # skip other eoln char
                    cursor += 1
            elif next in OCTALDIGITS:
                #print "octal1", ascii(next)
                octalstring = next
                if inlist[cursor] in OCTALDIGITS:
                    octalstring = octalstring+inlist[cursor]
                    #print "octal2", octalstring
                    cursor += 1
                    if inlist[cursor] in OCTALDIGITS:
                        octalstring = octalstring+inlist[cursor]
                        #print "octal3", octalstring
                        cursor += 1
                octal = int(octalstring, 8)
                thischar = chr(octal)
                #print cursor, len(outlist), "octal char", ascii(octalstring), "-->", ascii(thischar)
            else:
                raise ValueError("don't understand string escape sequence %s %s" % (ascii(thischar), ascii(next)))
        else:
            #print cursor, len(outlist), "passthru", ascii(thischar), "-->", ascii(thischar)
            pass # no translation needed
        outlist.append(thischar)
    result = ''.join(outlist)
    #print "UE", ascii(s), "-->", ascii(result)
    return result

def unescapePDFCharacterSequence(s):
    s = str(s) # in case it is an object, convert to string
    s0 = s[0]
    if s0=="(":
        return unescapePDFString(s)
    elif s0=="<":
        return unescapePDFText(s)
    else:
        raise ValueError("bad pdf char sequence starter %s" % ascii(s0))

def _unfilterStream(data,D):
    data = bytesFromNative(data)
    if not DECOMPRESSIMAGES and D.get('Subtype',None)==('name','Image'): return 1,data
    filters = originalfilters = D.get("Filter",None)
    if verbose: print('filters=%s' % ascii(filters))
    backout = 0
    if filters is not None:
        originaldata = data
        del D["Filter"]
        (ind, fx) = filters
        if ind == "array":
            filters = fx
        else:
            filters = [filters]
        # do decoding in order given
        for fm in filters:
            (n, f) = fm
            if n!="name":
                raise ValueError("filter not named %s" % ascii(fm))
            if f=="FlateDecode":
                if verbose>=5: print('FlateDecode')
                try:
                    data = pdfdoc.PDFZCompress.decode(data)
                    dP = D.get('DecodeParms',(None,{}))[1]
                    pred = dP.get("Predictor", (None,1))[1]
                    # pred 1 == no pred
                    if pred != 1:
                        nc = dP["Columns"][1]
                        # PNG prediction:
                        if pred >= 10 and pred <= 15:
                            out = [].extend
                            # PNG prediction can vary from row to row
                            rowlen = nc + 1
                            assert len(data) % rowlen == 0
                            prow = (0,) * rowlen
                            for r in range(len(data) / rowlen):
                                row = [ord(x) for x in data[(r*rowlen):((r+1)*rowlen)]]
                                fb = row[0]
                                if fb == 0:
                                    pass
                                elif fb == 1:
                                    for i in range(2, rowlen):
                                        row[i] = (row[i] + row[i-1]) % 256
                                elif fb == 2:
                                    for i in range(1, rowlen):
                                        row[i] = (row[i] + prow[i]) % 256
                                else:
                                    # unsupported PNG filter
                                    raise ValueError("Unsupported PNG filter %r" % fb)
                                prow = row
                                out([chr(x) for x in row[1:]])
                            data = ''.join(out.__self__)
                        else:
                            # unsupported pred
                            raise ValueError("Unsupported flatedecode predictor %r" % pred)
                except:
                    if verbose: print("WARNING: unable to zcompress.decode")
                    backout=1
                    break
            elif f=="ASCII85Decode":
                if verbose>=5: print('ASCII85Decode')
                try:
                    data = pdfdoc.PDFBase85Encode.decode(data)
                except:
                    if verbose: print("WARNING: unable to base 85 decode")
                    backout=1
                    break
            else:
                #raise ValueError("can't handle filter %s" % ascii(f))
                backout = 1
                break
        if backout:
            # for example got DCTdecode, which we don't do yet
            data = originaldata
            D["Filter"] = originalfilters
    if verbose>9: print('-->backout=%s data=%s' % (backout,ascii(data)))
    return backout, data

_extra_info_keys = ("MediaBox","CropBox","BleedBox","TrimBox","ArtBox", "Rotate")
class PDFFileCompilation:
    def __init__(self, prefix):
        self.prefix = prefix
        self.sequence = []
        # mapping name --> reference
        self.references = {}
        # mapping name --> tree
        self.trees = {}
        # mapping name --> pdfdoc object
        self.objects = {}
        # mapping name --> list of name
        self.uses = {}
        self.catalogName = None
        self.pageList = []
        self.lastObjectNumber = self.lastGenerationNumber = \
        self.encrypt = self.password = self.Ownerkey = self.IDarray1 = self.Permissions = None
        self.plainText = {}
    def encryptParameters(self, password, OwnerKey, UserKey, IDarray1, Permissions):
        #print 'encryptParameters', password
        if self.encrypt:
            raise ValueError("encryptParameters set twice (not permitted).")
        (self.password, self.Ownerkey, self.IDarray1, self.Permissions) = (
            password, OwnerKey, IDarray1, Permissions)
        from rlextra.utils import pdfencrypt
        self.encrypt = pdfencrypt.encryptionkey(password, OwnerKey, Permissions, IDarray1)
        pdfencrypt.checkU(self.encrypt, UserKey)
    def flattenAnnots(self,pann):
        pann = self.resolve(pann)
        if isinstance(pann, pdfdoc.PDFArray):
            R = []
            list(map(R.extend,[self.flattenAnnots(r) for r in pann.sequence]))
            return R
        else:
            return [pann]
    def pagesAsForm(self, pagenumbers, formprefix, all=None,fformname=None):
        """get dictionary of name-->object including form streams for the pagenumbers
           page numbers start at 0 and page 4 will have the pdfgen form
           name formprefix4 (if formprefix=="formprefix")
        """
        # result of function: object reference name --> pdfdoc object
        # special entry None --> list of EXTERNAL names of forms created (NOT INTERNALS!)
        nameToObject = {}
        nameToObject[None] = formNamesList = []
        #print "self.uses", self.uses
        import copy
        uses = nameToObject[0] = self.uses
        # collect all referents in pages (but don't trace parent, don't include pages themselves)
        requiredNames = {}
        forbiddenNames = {}
        # contents streams and arrays will have to be removed later, remember them.
        pageContentNames = {}
        pageAnnots = {}
        # if all is set then do all pages
        if all:
            pagenumbers = range(len(self.pageList))
        for n in pagenumbers:
            # collect all references in the page dictionaries...
            pagename = self.pageList[n]
            forbiddenNames[pagename] = pagename
            pagerefs = list(uses[pagename]) # make a copy for modification
            page = self.objects[pagename]
            if "Parent" in page.dict:
                raise ValueError("parent not removed")
            if "Annots" in page.dict:
                pageAnnots.setdefault(n,[]).extend(self.flattenAnnots(page.dict['Annots']))
                del page.dict['Annots']
            for name in pagerefs:
                requiredNames[name] = name
            compiledpagerep = self.trees[pagename]
            (ind, dict) = compiledpagerep
            if 'Annots' in dict:
                del dict['Annots']
            dict = dict.copy()
##            del dict["Parent"] # don't trace the parent link
##            for (ind2, thing) in dict.values():
##                if ind2=="R": # reference: trace it
##                    name = self.internalName(thing)
##                    requiredNames[name] = name
            # also, remember the references to content streams, and any indirect references to them
            # we don't need to include them twice in the output!
            c = dict.get("Contents",None)
            if c is not None:
                # resolve it if it is a reference
                while 1: # we already know there are no loops (previous steps)
                    (ind, thing) = c
                    if ind=="R":
                        c = thing
                        name = self.internalName(thing)
                        #print "name=", name
                        pageContentNames[name] = name
                    else:
                        break # stop at first non-reference
                # if content is an array, loop over contents
                (ind, thing) = c
                if ind=="array":
                    for (ind2, thing2) in thing:
                        if ind2=="R": # I assume the reference is direct... (some extra garbage included if not)
                            name = self.internalName(thing2)
                            #print "name=", name
                            pageContentNames[name] = name
        # transitively close the required references
        check = list(requiredNames.keys())
        while check:
            currentname = check[-1]
            del check[-1]
            requiredNames[currentname] = currentname
            for othername in uses[currentname]:
                if othername in forbiddenNames:
                    #print "forbidden", othername, "from", currentname
                    raise ValueError("forbidden object encountered "+ascii((othername, currentname)))
                if othername not in requiredNames:
                    check.append(othername)
        # now remove from the references names of page content arrays and streams
        # (these must be *traced* but should not be included in the output, since they are converted)
        for name in list(pageContentNames.keys()):
            del requiredNames[name]
        # store the required elements in result
        rk = list(requiredNames.keys())
        for name in rk:
            ob = self.objects[name]
            nameToObject[name] = ob
        # construct the forms
        for n in pagenumbers:
            externalFormName = fformname or formprefix+ascii(n)
            fformname = None
            formName = pdfdoc.xObjectName(externalFormName)
            formNamesList.append(externalFormName)
            pagename = self.pageList[n]
            # remember objects required for this form
            #print formName, "uses", uses[pagename]
            uses[formName] = list(uses[pagename])
            compiledpagerep = self.objects[pagename]
            pdict = compiledpagerep.dict
            # get the content for the page
            contents = pdict.get("Contents",None)
            if contents is not None:
                contents = self.resolve(contents)
                # there are two cases, single stream or array of stream
                filters = None
                if isinstance(contents, pdfdoc.PDFStream):
                    filters = contents.dictionary.dict.get("Filter", None)
                    contentstring = contents.content
                elif isinstance(contents, pdfdoc.PDFArray):
                    contentlist = []
                    for s in contents.sequence:
                        s = self.resolve(s)
                        # if it has an unexpanded filter give up!
                        if "Filter" in s.dictionary.dict:
                            raise ValueError("unexpanded filter in page array stream: giving up "+\
                                  ascii(s.dictionary.dict["Filter"]))
                        contentlist.append(s.content)
                    contentstring = ''.join(contentlist)
                else:
                    raise ValueError("bogus contents %s" % ascii(contents))
            else:
                contentstring = ''
                filters = None

            #RGB this is almost certainly not needed, the previous code was certainly wrong as it turned
            # \r\n into \r\r\n
            #if not COMPACT and not filters:
            #   contentstring = contentstring.replace("\r\n", "\n").replace("\r","\n").replace("\n","\r\n")
            self.plainText[n] = contentstring

            theform = pdfdoc.PDFStream()
            sdict = theform.dictionary
            # this attempt at security failed since it would require back patching the Resources Dict...
            # the idea was to add "evaluation copy" directly on top of the form.
##            if contentadjustment:
##                if filters:
##                    raise "Unpaid evaluation copy cannot handle filters: " + ascii(filters)
##                # mess up the form
##                contentstring = ("q\n%s\n" %contentstring)+contentadjustment
            # do compression
            if filters is not None:
                sdict["Filter"] = filters # HOW DID THAT GET COMMENTED?
                if verbose>1: print("filters is %s" % ascii(filters))
                pass
            elif COMPACT:
                if verbose>5: print("FlateDecode set")
                sdict["Filter"] = pdfdoc.PDFName("FlateDecode")
                contentstring = pdfdoc.PDFZCompress.encode(contentstring)
            theform.content = contentstring
            sdict["Type"] = pdfdoc.PDFName("XObject")
            sdict["Subtype"] = pdfdoc.PDFName("Form")
            sdict["FormType"] = 1
            bbox = sdict["BBox"] = pdict["MediaBox"]
            matrix = [1,0,0,1,0,0]
            deg = pdict.get('Rotate',0)
            if deg!=0:
                #hackish attempt to fix up for rotations other than zero
                #note that this assumes the existing matrix is [1,0,0,1,0,0]
                bb =  bbox.sequence
                matrix = inverse(rotate(deg))
                #pts = where the bbox corners will end up; we ignore anything negative as that has already disappeared
                bb = bb[:2], bb[2:], [bb[0],bb[3]], [bb[2],bb[1]]
                pts = [zTransformPoint(matrix,(max(0,x),max(0,y))) for x, y in bb]

                #set the translation required to bring the box back into view
                matrix = list(matrix)
                matrix[-2] = max(0,-min([x[0] for x in pts]))
                matrix[-1] = max(0,-min([x[1] for x in pts]))
                sdict["Rotate"] = 0
            sdict["Matrix"] = pdfdoc.PDFArray(matrix) # XXXX IS THIS ALWAYS RIGHT???
            sdict["Resources"] = pdict["Resources"]

            #copy extra information eg boxes, rotate
            for k in _extra_info_keys:
                if k in pdict:
                    sdict[k] = pdict[k]

            nameToObject[formName] = theform
        if pageAnnots:
            for V in pageAnnots.values():
                for v in V:
                    v.normalize()
            nameToObject['pageAnnots'] = pageAnnots

        return nameToObject
    def getReference(self, idAndVersion):
        name = self.internalName(idAndVersion)
        return self.references[name]
    def doTranslations(self, catalogInfo):
        "translate all elements to pdfdoc structures, after all names have been added"
        for name in self.sequence:
            #print "NAME IS", name
            #stop
            (prefix, pair) = name
            (self.lastObjectNumber, self.lastGenerationNumber) = pair # for encryption
            tree = self.trees[name]
            object = self.translateTree(tree)
            self.objects[name] = object
        self.catalogName = self.internalName(catalogInfo)
    def sanitizePages(self, save=("Type", "Contents","Resources")+_extra_info_keys, saveAnnots=False):
        "for all page dictionaries, cut out attributes not saved (for form translation)"
        #print "sanitizing"
        if saveAnnots:
            save = save + ('Annots',)
        for name in self.sequence:
            tree = self.trees[name]
            (ind, val) = tree
            if ind=="dictionary":
                (ind2, typ) = val.get("Type", (None, None))
                #print " dictionary", `typ`
                if typ=="Page":
                    for name in list(val.keys()):
                        if name not in save:
                            del val[name]
    def resolve(self, ob):
        "find object referred to if reference (after translations)"
        seen = {}
        #ob = self.objects[name]
        while isinstance(ob, pdfdoc.PDFObjectReference):
            name = ob.name
            if name in seen:
                raise ValueError("infinite reference loop %s" % ascii(name))
            seen[name] = 1
            ob = self.objects[name]
        return ob
    def populatePageList(self):
        "create list of names for pages self.pageList, after doTranslations"
        catalog = self.objects[self.catalogName]
        # note catalog is a pdfdoc.pdfDictionary, not a pdfdoc.pdfcatalog
        pagesref = catalog.dict["Pages"]
        pagesname = pagesref.name
        pages = self.resolve(pagesref)
        self.expandPages(pagesref)
    def expandPages(self, pagesref, inheritancedict=None,
                    #inheritableattributes = ("MediaBox", "Resources", "CropBox", "Rotate", "Dur", "Hid", "Trans")
                    # for these purposes we are only interested in mediabox and resources (others may mess us up)
                    inheritableattributes = ("MediaBox", "Resources")
                    ):
        # note: to make pageToForm conversion work, attributes "inherited" down
        # the Pages tree are explicitly copied down and references are patched.
        ### NOTE: THIS MAKES THE PAGE DICTIONARY PARSE TREES NO LONGER IN SYNC WITH THE
        ###  THE COMPILED PDFDOC DICTIONARY OBJECTS
        if inheritancedict is None:
            inheritancedict={}
        else:
            inheritancedict = inheritancedict.copy() # no sharing across recursions
        pagesName = pagesref.name
        pagesDictionary = self.resolve(pagesref)
        # paste in inheritable attributes if absent (so page dicts are self contained)
        pd = pagesDictionary.dict
        for a in inheritableattributes:
            if a in inheritancedict:
                if not (a in pd): # force down the inherited value
                    inheritedobject = inheritancedict[a]
                    pd[a] = inheritedobject
                    # patch the uses list!
                    if isNonPrimitiveInstance(inheritedobject):
                        # add any references in the inherited parse tree to the uses list for page dictionary
                        self.findReferences(pagesName, inheritedobject.__parse_tree__)
            elif a in pd:
                inheritancedict[a] = pd[a] # remember this inherited value
        dictType = pagesDictionary.dict["Type"]
        if dictType==PAGENAME:
            # it is a page. record its name
            self.pageList.append(pagesName)
        elif dictType==PAGESNAME:
            # it is a pages tree: expand it
            kids = self.resolve(pagesDictionary.dict["Kids"]) # should be an array
            for p in kids.sequence:
                self.expandPages(p, inheritancedict)
    def translateTree(self, tree, allowNULL=0):
        #dieHere
        # Note: decryption herein will "decode" the encryption dictionary strings.
        #   turning them into garbage (since they should not be decrypted).  But this doesn't
        #   matter, since they have already been extracted in order to enable encryption....
        (indicator, val) = tree
        if indicator=="R":
            name = self.internalName(val)
            try:
                result = self.references[name]
            except KeyError:
                if allowNULL: return NULL
                raise
        elif indicator in ("true", "false", "null", "number"):
            result = val
        elif indicator=="text":
            if not self.encrypt:
                #result = "<%s>" % val # note: hexification was not undone in this parser ! BUG!
                value = unescapePDFText("<%s>" % val)
                result = pdfdoc.PDFText(value)
            else:
                value = unescapePDFText("<%s>" % val)
                from rlextra.utils import pdfencrypt
                val2 =  pdfencrypt.encodePDF(self.encrypt, self.lastObjectNumber, self.lastGenerationNumber, value)
                result = pdfdoc.PDFText(val2)
        elif indicator=="string":
            if not self.encrypt:
                # must decode and reencode in case of encryption (?)
                value = unescapePDFString("(%s)" % val)
                result = pdfdoc.PDFString(value)
            else:
                value = unescapePDFString("(%s)" % val)
                from rlextra.utils import pdfencrypt
                val2 =  pdfencrypt.encodePDF(self.encrypt, self.lastObjectNumber, self.lastGenerationNumber, value)
                #print "before", ascii(val)
                #print "after", ascii(val2)
                result = pdfdoc.PDFString(val2)
                #diehereString WTF????
        elif indicator=="name":
            result = pdfdoc.PDFName(val)
        elif indicator=="array":
            elts = []
            for v in val:
                e = self.translateTree(v)
                elts.append(e)
            result = pdfdoc.PDFArray(elts)
        elif indicator=="dictionary":
            d = {}
            for k in list(val.keys()):
                v = val[k]
                e = self.translateTree(v,allowNULL=1)
                if e is not NULL: d[k] = e
            result = pdfdoc.PDFDictionary(d)
        elif indicator=="stream":
            (dict, content) = val
            test = len(content)
            result = pdfdoc.PDFStream()
            ### decode content here, if desired...
            if self.encrypt:
                from rlextra.utils import pdfencrypt
                content2 =  pdfencrypt.encodePDF(self.encrypt, self.lastObjectNumber, self.lastGenerationNumber, content)
                # debug here...
                content = content2
                #diehereStream
            if DECOMPRESSSTREAMS:
                backout, content = _unfilterStream(content,dict)
            if verbose>=9: print("streamcontent=%s" % ascii(content))
            dicttuple = ("dictionary", dict)
            pd = self.translateTree(dicttuple)
            result.dictionary = pd
            result.content = content
            #### debugging!!!
            #result.content = ""
        else:
            raise ValueError("don't understand %s indicator" % ascii(indicator))
        if isNonPrimitiveInstance(result):
            # STORE THE PARSE TREE
            result.__parse_tree__ = tree
        return result
    def internalName(self, externalName):
        return (self.prefix, externalName)
    def addParse(self, idAndVersion, tree):
        name = self.internalName(idAndVersion)
        sequence = self.sequence
        if name in sequence:
            raise ValueError("name not new %s" % ascii(name))
        self.sequence.append(name)
        self.uses[name] = []
        self.trees[name] = tree
        self.references[name] = pdfdoc.PDFObjectReference(name)
        self.uses[name] = []
        #if findrefs:
        #    self.findReferences(name, tree)
        # self.object translated in later pass.
    def findAllReferences(self):
        sequence = self.sequence
        trees = self.trees
        for name in sequence:
            tree = trees[name]
            self.findReferences(name, tree)
    def findReferences(self, name, tree):
        (indicator, val) = tree
        if indicator in ("true", "false", "null", "number", "text", "name", "string"):
            pass # no references in these
        elif indicator=="R":
            # found a reference!
            rname = self.internalName(val)
            uses = self.uses[name]
            if rname not in uses:
                uses.append(rname)
        elif indicator=="array":
            for v in val:
                self.findReferences(name, v)
        elif indicator=="dictionary":
            for v in list(val.values()):
                self.findReferences(name, v)
        elif indicator=="stream":
            (dict, content) = val
            for v in list(dict.values()):
                self.findReferences(name, v)
        else:
            raise ValueError("don't understand indicator "+ascii(indicator))

def _pdf_escaped(text,c):
    '''return 1 if text[c] is preceded by an odd number of backslash chars'''
    for i in range(c-1,-1,-1):
        if text[i]!='\\':
            return (c-i+1)%2
    return c%2

class PDFParseContext:
    encrypt = None

    # class global dispatch dictionary for first char parse lookup...
    dispatch_dictionary = {}

    def __init__(self, text, prefix="no prefix"):
        self.compilation = PDFFileCompilation(prefix)
        self.cursor = 0
        self.stack = []
        if not isNative(text): text = nativeFromBytes(text)
        self.text = text
        self.splittext = text.split()
        self.splitcursor = 0
        self.objectdict = {} # for collecting objects (number, version)-->object
        self.objectoffsets = {} # for collecting byte offests (number, version)-->object
        self.align()
        self.trailer = None

    def parse(self):
        try:
            self.innerParse()
        except:
            ot,ov,otb = sys.exc_info()[:]
            try:
                from rlextra.thirdparty.pyPdf import PdfFileWriter, PdfFileReader
            except ImportError:
                rl_reraise(ot,ov,otb)
            except:
                t,v,tb=sys.exc_info()
                oerr = getStringIO()
                traceback.print_exception(ot,ov,otb,None,oerr)
                del ot, ov, otb
                oerr = ('\n'+oerr.getvalue().strip()).replace('\n','\n    ')+'\n'
                v = t('%s\n\noriginal error\n%s' % (''.join(map(str,v.args)),oerr))
                rl_reraise(t,v,tb)
            try:
                input = PdfFileReader(getBytesIO(asBytes(self.text,'latin1')))
                output = PdfFileWriter()
                n = input.getNumPages()
                for i in range(n):
                    output.addPage(input.getPage(i))
                del input
                f=getBytesIO()
                output.write(f)
                self.__init__(f.getvalue(),self.compilation.prefix) #full reinitialize
                del f
                self.innerParse()
            except:
                t,v,tb=sys.exc_info()
                oerr = getStringIO()
                traceback.print_exception(ot,ov,otb,None,oerr)
                del ot, ov, otb
                oerr = ('\n'+oerr.getvalue().strip()).replace('\n','\n    ')+'\n'
                v = t('%s\n\noriginal error\n%s' % (''.join([asNative(_v,'latin1') for _v in v.args]),asNative(oerr,'latin1')))
                rl_reraise(t,v,tb)

    def innerParse(self):
        """parse the file by looking at the end, then the cross reference, etc.
        """
        text = self.text
        # look for the xref
        endoftext = text[-50:]
        s = "startxref"
        f = text.rfind(s)
        if f<0: raise ValueError("can't find %s at end of file %s" % (ascii(s), ascii(endoftext)))
        f = f+len(s)
        f2 = text.find("%%EOF", f)
        if f2<0: raise ValueError("eof marker not found at file end %s" % ascii(endoftext))
        startxref = int(text[f:f2].strip())
        catalog = None
        # repeatedly parse all cross reference tables
        while startxref is not None:
            #print "parsing xref at", startxref
            self.cursor = startxref
            self.splitcursor = 0
            self.cursorcheck()
            if self.cursor!= startxref:
                raise ValueError("alignment error")
            self.splitcursor += 1 # move past "xref" keyword
            self.cursorcheck()
            # parse the entries
            self.parse_xref_entries()
            #print self.objectoffsets
            # now parse the trailer
            if self.splittext[self.splitcursor]!="trailer":
                raise ValueError("trailer not found after xref %s" % self.cursor)
            self.splitcursor += 1 # move past "trailer"
            self.cursorcheck()
            (marker, trailerdict) = self.getdictionary()
            # fail if encryption is specified
            if "Encrypt" in trailerdict:
                if self.encrypt is not None:
                    raise ValueError("more than one encryption dictionary found")
                self.encrypt = trailerdict["Encrypt"]

            if "ID" in trailerdict:
                # remember the id (needed for decryption)
                self.PDFID = trailerdict["ID"]
            # do we need to repeat?
            if "Prev" in trailerdict:
                (mark, startxref) = trailerdict["Prev"]
            else:
                startxref = None # no more xrefs
            if catalog is None:
                self.catalog = catalog = trailerdict["Root"] # object reference
                self.trailerdict = trailerdict
            # XXXX for now ignore other info in trailerdict
        # now parse ALL objects (even ones that can't be used anymore, possibly)
        objectoffsets = self.objectoffsets.copy()
        lastlen = len(objectoffsets)
        dd = self.dispatch_dictionary
        # THIS LOOP MUST REPEAT UNTIL ALL OBJECT REFS CAN RESOLVE (REFERENCES MAY NOT BE IN FILE ORDER)
        while objectoffsets:
            #print "looking for", lastlen, "objects"
            keys = list(objectoffsets.keys())
            keys.sort()
            for k in keys:
                #print "trying to resolve", k
                offset = objectoffsets[k]
                self.cursor = offset
                self.splitcursor = 0
                self.cursorcheck()
                stack = self.stack = []
                # first find x y obj
                for n in (0,1):
                    numinfo = self.getnumber()
                    stack.append(numinfo)
                    self.skipwhitespace()
                idinfo = self.getobjectID()
                stack.append(idinfo)
                self.skipwhitespace()
                # now find some object (maybe fail)
                c = self.cursor
                firstchr = text[c]
                unboundparsemethod = dd.get(firstchr, None)
                if unboundparsemethod is None:
                    raise ValueError("can't parse at %s %s" % (c, ascii(text[c:c+20])))
                failed = 0
                try:
                    parsepair = unboundparsemethod(self)
                except UnresolvedRefError:
                    #print "resolution failed", k
                    failed = 1 # need to try again for this one
                else:
                    #print "::: got :::", idinfo
                    #pprint(parsepair)
                    stack.append(parsepair)
                    self.skipwhitespace()
                    # are we at a stream keyword
                    if text[self.cursor]=="s":
                        try:
                            streaminfo = self.getstream()
                        except UnresolvedRefError:
                            #print "resolution failed", k
                            failed = 1 # need to try again for this one
                        else:
                            stack.append(streaminfo)
                # now we should be at the endobject keyword
                if not failed:
                    self.skipwhitespace()
                    self.getindirectObject() # this records the object in self
                    # we're done with this one!
                    #print "resolved", k
                    del objectoffsets[k]
            newlen = len(objectoffsets)
            if newlen==lastlen: raise ValueError("no progress resolving objects %s" % ascii(list(objectoffsets.keys())))
            lastlen = newlen
        # XXXX NOW ARRANGE THINGS AS IF IT HAD BEEN DONE BY THE PROCESS METHOD
        #self.objectdict[pair] = object
        od = self.objectdict
        keys = list(od.keys())
        keys.sort
        stack = []
        iddict = {}
        for k in keys:
            (id, version) = k
            #if id in iddict: raise ValueError("multiple object versions not yet supported... %s" % id)
            #iddict[id]=id
            ob = od[k]
            stack.append(ob)
        stack.append(("trailer", self.trailerdict))
        self.stack = stack
        return stack

    def parse_xref_entries(self):
        "parse xref entries starting at cursor, add to self.objectoffsets (cursor *after* xref keyword)"
        sc = self.splitcursor
        text = self.text
        splittext = self.splittext
        # HACK motivated by acrobat: if we are at the startxref keyword then realign at the xref
        # (problem observed by arw)
        thispoint = splittext[sc]
        #print "parse_xref_entries at", splittext[sc:sc+4]
        if thispoint=="startxref":
            # find where the xref REALLY STARTS!
            newcursor = int(splittext[sc+1])
            # go there
            self.cursor = newcursor
            self.splitcursor = 0
            self.cursorcheck()
            sc = self.splitcursor+1
        done = None
        objectoffsets = self.objectoffsets
        if splittext[sc+1]=='obj':
            #it's likely we're in a new style stream crossref situation
            try:
                self.cursor = self.splitindices[sc+2]
                dummy,d = self.getdictionary()
                t=d['Type'][1]
                if t!='XRef': raise ValueError('Incorrect dictionary type %r parsing new style xref' % t)
                self.stack = [(dummy,d)]
                dummy,(d,content) = self.getstream()
                backout, content = _unfilterStream(content,d)
                #print backout,content
                raise ValueError('cannot continue at present with new style xref')
            except:
                t,v,tb = sys.exc_info()
                s = 'Failed to get stream XRef\n%s' % '\n'.join(map(str,v.args))
                rl_reraise(t,t(s),tb)
        while done is None:
            # get section head info
            #sc = self.splitcursor
            #print "section head at", splittext[sc:sc+6]
            firstentrynumber = int(splittext[sc])
            nentries = int(splittext[sc+1])
            sc = self.splitcursor = sc+2
            #self.cursorcheck()
            entrynumber = firstentrynumber
            #print "nentries", nentries
            for i in range(nentries):
                # get this entry (don't check eoln requirements too carefully)
                #print "near", splittext[sc:sc+5]
                offset, generation, indicator = splittext[sc:sc+3]
                sc = sc+3
                #print "advanced to", splittext[sc: sc+3]
                offset = int(offset, 10) # decimal! not octal!
                generation = int(generation, 10)

                #workaround for a bug in matplotli's PDF backend who have an 'n' on the first row
                if offset == 0 and generation == 65535 and indicator == 'n':
                    indicator = 'f'

                if indicator == 'n':
                    # in use: record it
                    objectkey = (entrynumber, generation)
                    if objectkey in objectoffsets: # apparently this is not an error!
                        #raise ValueError("already defined %s" %ascii(objectkey))
                        pass # ignore it: use the "more recent" value which occurs later in the file (for acrobat written files)
                    else:
                        objectoffsets[ objectkey ] = offset
                elif indicator=='f':
                    # free: ignore it
                    pass
                else:
                    raise ValueError("bad refentry indicator %s" % ascii(indicator))
                entrynumber += 1
            # are we done?
            try:
                test = int(splittext[sc])
                #print "not done near", splittext[sc:sc+6]
            except:
                done = 1 # yep!
                #print "done near", splittext[sc:sc+6]
        self.splitcursor = sc
        self.cursorcheck()

    def resolve(self, thing):
        """resolve thing, or signal unresolved."""
        od = self.objectdict
        limit = len(od)
        count = 0
        #print "thing", thing
        (mark, rep) = thing
        while mark=="R":
            count += 1
            if count>limit+2:
                raise ValueError("infinite object loop "+ascii(thing))
            test = od.get(rep,None)
            if test is None:
                raise UnresolvedRefError(rep)# pass the rep up as the value
            thing = (mark, rep) = test
        return thing

    def getindirectObject(self):
        t = self.text; c = self.cursor
        test = self.gettrue("endobj")
        # pop the obid and object off the stack
        stack = self.stack
        [(m1,pair), object] = stack[-2:]
        if m1!="obj":
            raise ValueError("expect objectID before object before 'endobj' got %s  at %s %s" % (m1, c, ascii(t[c:c+20])))
        del stack[-2:]
        self.objectdict[pair] = object
        # record in compilation
        self.compilation.addParse(pair, object)
        return ("indirectobject", (pair, object))

    dispatch_dictionary["e"] = getindirectObject

    def getobjectref(self, keyword="R"):
        t = self.text; c = self.cursor
        test = self.gettrue(thing=keyword)
        # pop off the two numbers from the stack
        stack = self.stack
        [(m1,n1), (m2,n2)] = stack[-2:]
        if isinstance(n1,int) and isinstance(n2,int):
            del stack[-2:]
            return (keyword, (n1,n2))
        raise ValueError("expected 2 ints before %s %s %s" % (keyword, c, ascii(t[c:c+20])))

    def getobjectID(self):
        return self.getobjectref(keyword="obj")

    dispatch_dictionary["R"] = getobjectref
    dispatch_dictionary["o"] = getobjectID

    def gettrue(self, thing="true"):
        lt = len(thing)
        t = self.text; c = self.cursor
        endkw = c+lt
        if t[c:endkw]!=thing:
            raise ValueError("`%s` keyword not found %s %s" % (thing, c, ascii(t[c:c+20])))
        self.cursor = endkw
        self.cursorcheck()
        return (thing, thing)

    dispatch_dictionary["t"] = gettrue

    def getfalse(self):
        return self.gettrue(thing="false")

    dispatch_dictionary["f"] = getfalse

    def getnull(self):
        return self.gettrue(thing="null")

    dispatch_dictionary["n"] = getnull

    def getstream(self):
        t = self.text; c = self.cursor
        LF = "\n"; CR = "\r"
        if t[c: c+6]!="stream":
            raise ValueError("not at stream keyword at %s %s" % (c, ascii(t[c:c+20])))
        if t[c+6] == LF:
            streamstart = c+7
        elif t[c+6]==CR and t[c+7]==LF:
            streamstart = c+8
        else:
            raise ValueError("invalid stream keyword end of line %s %s" % (c, ascii(t[c:c+20])))
        # pop the dictionary off the stack
        stack = self.stack
        (mark, dict) = stack[-1]
        if mark!="dictionary":
            raise ValueError("stream not preceded by dictionary %s %s %s" %(mark, c, ascii(t[c:c+20])))
        del stack[-1]
        # find the length
        streamlen = dict.get("Length", None)
        if streamlen is None:
            raise ValueError("stream with no length key %s %s" % (c, ascii(t[c:c+20])))
        # streamlen may be an indirect object!
        streamlen = self.resolve(streamlen)
        (tag, slen) = streamlen
        streamend = streamstart+slen
        streamcontent = t[streamstart:streamend]
        # should be at "endstream"
        if t[streamend: streamend+9]!="endstream":
            if verbose: print("EXPECTED STREAM LENGTH %s" % slen)
            #print "text length", len(text)
            # try to tolerate whitespace, with a warning
            if verbose: print("WARNING stream length doesn't end at endstream keyword %s %s" % (
                                  streamend, ascii(t[streamend:streamend+20])))
            endstreamkw = "endstream"
            lkw = len(endstreamkw)
            f = t.find(endstreamkw, streamend)
            if f<0:
               pass
            elif t[streamend:f].strip()=="":
               streamend = f
            streamend2 = streamend+lkw
            if t[streamend: streamend2]!="endstream":
                raise ValueError("stream length doesn't end at endstream keyword %s %s" % (
                                  streamend, ascii(t[streamend:streamend+20])))
        self.cursor = streamend+9
        self.cursorcheck()
        return ("stream", (dict, streamcontent))

    dispatch_dictionary["s"] = getstream

    def getdictionary(self):
        t = self.text; c = self.cursor
        (dummy, stack) = self.getarray(startmark="<<", endmark=">>")
        if len(stack)%2:
            #print stack
            raise ValueError("odd number of elements in dictionary contents %s %s" %(c, t[c:c+20]))
        stackcopy = list(stack)
        D = {}
        while stackcopy:
            [k,v] = stackcopy[:2]
            del stackcopy[:2]
            (kmark, kname) = k
            # strip out the marker, key assumed to always be a name
            if kmark!="name":
                raise ValueError("dictionary key must be name %s %s %s" % (kmark, c, ascii(t[c:c+20])))
            D[kname] = v
        return ("dictionary", D)

    # getdictionary called by dispatchanglebracket below

    def skiptokeyword(self, keyword):
        # does this handle (missing) whitespace correctly always?
        done = None
        sc = self.splitcursor
        st = self.splittext
        lk = len(keyword)
        while done is None:
            sc += 1
            ss = st[sc]
            if ss[:lk]==keyword: done = 1
        self.splitcursor = sc = sc+1
        self.cursor = self.splitindices[sc]

    def skiptostring(self, st):
        #print "skiptostring", st
        t = self.text
        c = self.cursor
        c2 = t.find(st, c)
        if c2<0: raise ValueError("can't find %s past %s" % (ascii(st), c))
        self.cursor = c2+len(st) # actually skip PAST!
        self.cursorcheck()

    def getarray(self, startmark="[", endmark="]"):
        start = c = self.cursor
        text = self.text
        if startmark and text[c:c+len(startmark)]!=startmark:
            raise ValueError("getname not starting at %s %s %s"%(ascii(startmark), c, ascii(text[c:c+20])))
        if startmark:
            self.cursor = c+len(startmark)
        save_stack = self.stack
        stack = self.stack = []
        done = None
        ltext = len(text)
        lendmark = 1
        if endmark:
            lendmark = len(endmark)
        dd = self.dispatch_dictionary
        while done is None:
            self.skipwhitespace()
            c = self.cursor
            if verbose>1: print("scanning array %s" % ascii(text[c:c+20]))
            if c>ltext:
                if endmark:
                    raise ValueError("unclosed array at %s %s" % (start, ascii(text[start:start+20])))
                self.cursor = ltext
                done = 1
            elif text[c:c+lendmark]==endmark:
                if verbose>2: print(ascii(endmark)+" endmark found")
                self.cursor = c+lendmark
                done = 1
            else:
                # dispatch
                firstchr = text[c]
                unboundparsemethod = dd.get(firstchr, None)
                if unboundparsemethod is None:
                    linecount = len(text[:c].split("\n"))
                    raise ValueError("can't parse at line %s %s %s%s" % (
                        linecount,  firstchr,ascii(text[c-50:c]),ascii(text[c:c+50]),))
                parsepair = unboundparsemethod(self)
                stack.append(parsepair)
                if verbose>1:
                    print("%s at %s" %(ascii(parsepair[0]), len(stack)))
        sequence = self.stack
        self.stack = save_stack
        self.cursorcheck()
        return ("array", sequence)

    dispatch_dictionary["["] = getarray

    def getnumber(self, text=None, cursor=None):
        """get a number at cursor or error"""
        # really simple implementation.. probably should use re instead...
        if text is None:
            text = self.text
        if cursor is None:
            cursor = self.cursor
        startnum = endnum = cursor
        ltext = len(text)
        while endnum<ltext and text[endnum] in "+-0123456789.":
            endnum += 1
        if startnum==endnum:
            raise ValueError("bad call to getnumber at %s %s" % (cursor, ascii(text[cursor:cursor+20])))
        numtext = text[startnum:endnum]
        try:
            num = int(numtext)
        except:
            num = float(numtext)
        self.cursor = endnum
        self.cursorcheck()
        return ("number", num)

    for c in "+-0123456789.":
        dispatch_dictionary[c] = getnumber

    def getstring(self, text=None, cursor=None, startmark="(", closemark=")"):
        """find a string in parens at cursor, or error
           NOTE: does not mess with content (yet) XXX???"""
        # start/closemark must have length 1 in this implementation!!!
        #print "getstring at", self.cursor
        ### in IRS form f1000.pdf I find the string (Form 1000 (Rev. August 1997))
        ### SO: if you find an open ( not preceded by \, must look for matching unescaped )
        ### I guess (this is undocumented, or at least poorly documented!)
        if cursor is None:
            c = self.cursor
        else:
            c = cursor
        if text is None:
            text = self.text
        if text[c]!=startmark:
            raise ValueError("getstring not starting at %s %s %s"%(ascii(startmark), c, ascii(text[c:c+20])))
        #cursor = c
        cursor = startstring = c+1
        # look for close paren not preceded by '\'
        #closed = None
        depth = 1
        while depth>0:
            #print "at", ascii(text[cursor:cursor+40]), depth
            closed = 0
            closestring = cursor
            # look for a close
            while not closed:
                closestring = text.find(closemark, closestring)
                if closestring<0:
                    raise ValueError("unclosed %s at %s %s"%(ascii(startmark), c, ascii(text[c:c+20])))
                if not _pdf_escaped(text,closestring):
                    closed = 1
                    depth -= depth
                else:
                    closestring += 1
            #print "closestring", closestring, `text[closestring: closestring+10]`
            #depth -= 1
            # BUT increment for every unescaped ( before it!
            foundallopens = 0
            openstring = cursor
            while not foundallopens:
                openstring = text.find(startmark, openstring)
                if openstring<0 or openstring>=closestring:
                    foundallopens = 1
                elif not _pdf_escaped(text,openstring):
                    #print "openstring", openstring, `text[openstring: openstring+10]`
                    depth += 1
                openstring += 1 # keep looking until all found
            cursor = closestring
            if depth>0:
                cursor += 1 # look to next
        s = text[startstring:closestring] # string left in PDF escaped format
        #print open("text.txt", "w").write(s)
        self.cursor = closestring+1
        self.cursorcheck()
        #print "string", s
        return ("string", s)

    dispatch_dictionary["("] = getstring

    def getname(self, text=None, cursor=None):
        "find /Name at cursor or error.  Do convert the #30 encoded hex characters"
        # really stupid implementation for now (no point in optimizing [yet]
        startmark = "/"
        if cursor is None:
            c = self.cursor
        if text is None:
            text = self.text
        if text[c]!=startmark:
            raise ValueError("getname not starting at %s %s %s"%(ascii(startmark), c, ascii(text[c:c+20])))
        startname = c+1
        endname = startname
        thischar = text[endname]
        while 0x21<=ord(thischar)<=0x7e and thischar not in "%()<>{}[]/":
            endname += 1
            if endname>len(text): break # should be an error?
            thischar = text[endname]
        if endname==startname:
            if verbose: print("empty name at %s %s" % (c, ascii(text[c:c+20])))
        s = text[startname: endname]
        # advance cursor
        self.cursor = endname
        self.cursorcheck()
        # look for hex chars #20
        index = 0
        while index<len(s): # len(s) may change!
            if s[index]=="#":
                hexchar = s[index+1: index+3]
                try:
                    hexord = int(hexchar, 16)
                    char = int2Byte(hexord)
                    s = "%s%s%s" %(s[:index], char, s[index+3:])
                except:
                    # this wasn't 2 digits after the hash, so wasn't a number
                    # so let it through.
                    pass
            index += 1
        return ("name", s)

    dispatch_dictionary["/"] = getname

    def gettext(self,text=None,cursor=None):
        """find text in <angle brackets> at cursor, or error, don't unhex the contents (yet)"""
        #if cursor is None: cursor = self.cursor
        #raise ValueError("gettext %s" % ascii(self.text[cursor: cursor+20]))
        (dummy, s) = self.getstring(text=None, cursor=None, startmark="<", closemark=">")
        #print "found text", s
        return ("text", s)

    #dispatch_dictionary["<"] = gettext
    def dispatchanglebracket(self):
        c = self.cursor
        t = self.text
        if t[c]!="<":
            raise ValueError("didn't find < at "+ascii(t[c:c+20]))
        if t[c:c+2]=="<<":
            return self.getdictionary()
        else:
            return self.gettext()

    dispatch_dictionary["<"] = dispatchanglebracket

    aligned = None

    def align(self):
        """find indices for starts of splitted text"""
        splittext = self.splittext
        text = self.text
        splitindices = self.splitindices = splittext[:]
        cursor = i = 0
        for st in splittext:
            cursor = text.find(st, cursor)
            splitindices[i] = cursor
            cursor = cursor + len(st)
            i += 1
        splitindices.append(len(self.text))
        self.aligned = 1
        self.cursor = 0
        self.splitcursor = 0
        # invariant:
        #  self.splitindices[self.splitcursor]<=self.cursor<=self.splitindices[self.splitcursor+1]
        #  self.cursor<=len(self.text) only = when done

    def cursorcheck(self):
        "make sure self.cursor and self.splitcursor refer to same virtual point in document"
        # actually, splitcursor can refer to the end of last whitespace *before* cursor
        if not self.aligned:
            if verbose:
                print("warning: cannot cursorcheck before alignment")
        cursor = self.cursor
        splitcursor = self.splitcursor
        if splitcursor>=len(self.splittext): # eof case
             self.cursor = len(self.text)
             self.splitcursor = len(self.splittext)
             return
        splitindices = self.splitindices
        [thisindex,nextindex] = splitindices[splitcursor:splitcursor+2]
        if cursor>=nextindex:
            # adjust the splitcursor forward
            from bisect import bisect
            self.splitcursor = bisect(splitindices, cursor)-1
            return
        if cursor<thisindex:
            self.cursor = thisindex
            return

    def skipwhitespace(self, whitespace=string.whitespace):
        c0 = c = self.cursor
        text = self.text
        #print "skipwhite at", `text[c:c+20]`
        splitindices = self.splitindices
        aligned = self.aligned
        splitcursor = self.splitcursor
        done = None
        count = 1
        while done is None:
            #print "skipwhite", count
            count += 1
            try:
                cc = text[c]
            except:
                raise ValueError("skip white to eof starting at %s" % c0)
            if cc=="%":
                # in comment: skip to eoln
                #print "comment at", `text[c:c+20]`
                c2 = text.find("\n", c)
                if c2<=0:
                    c2 = text.find("\r", c)
                if c2<=0:
                    raise ValueError("comment not closed by newline at %s" % c)
                c = c2+1
                while splitindices[splitcursor]<c:
                    splitcursor += 1
                c = splitindices[splitcursor]
                #print "skipped to", `text[c:c+20]`
            # treat startxref...%%EOF as whitespace too
            elif cc=="s" and text[c:c+9]=="startxref":
                self.skiptostring("%%EOF")
                c = self.cursor
                splitcursor = self.splitcursor
                #print "skipped startxref..%%EOF from", c0, "to", c, "total", c-c0
            elif cc in whitespace:
                # skip to next nonwhite
                splitcursor += 1
                c = splitindices[splitcursor]
            else:
                done = 1
        self.splitcursor = splitcursor
        self.cursor = c

def rewrite(frompdffile, topdffile):
    print("\n\n reformatting %s into %s " % (ascii(frompdffile), ascii(topdffile)))
    text = nativeFromBytes(open_and_read(frompdffile))
    p = PDFParseContext(text, frompdffile)
    p.parse()
    (ind, catinfo) = p.catalog
    c = p.compilation
    c.findAllReferences()
    catalog = c.getReference(catinfo)
    c.doTranslations(catinfo)
    doc = pdfdoc.PDFDocument()
    doc.Catalog = catalog
    # put all elements into file in order found
    print("\n\n\n Now installing objects in document")
    c.sequence.reverse() # REVERSE SEQUENCE (TESTING)
    for name in c.sequence:
        ob = c.objects[name]
        print("%s %s" % (ascii(name),ascii(ob)))
        doc.Reference(ob, name)
    # now write the file
    print("\n\nnow writing %s" % ascii(topdffile))
    doc.SaveToFile(topdffile, None) # no canvas needed, no outline
    print("rewrite done")

def combinePdfsInMemory(list_of_pdfs):
    """Take a list of PDf files 'in memory' and combine.

    Each item in the list is a PDF file held in memory as
    a string. This will use PageCatcher techniques to make
    one PDF file.
    """
    if len(list_of_pdfs)>1:
        buf = getBytesIO()
        canv = canvas.Canvas(buf)
        for pdfContent in list_of_pdfs:
            copyPagesInMemory(pdfContent, canv)
        canv.save()
        return buf.getvalue()
    else:
        try:
            return list_of_pdfs[0]
        except IndexError:
            raise ValueError('empty list passed to combinePdfsInMemory')

def combinePdfs(combinedFileName, list_of_pdf_filenames):
    "Merge list of PDF files into one.  Uses lots of memory :-)"
    list_of_pdf_content = []
    for filename in list_of_pdf_filenames:
        list_of_pdf_content.append(nativeFromBytes(open_and_read(filename)))
    bigPdf = combinePdfsInMemory(list_of_pdf_content)
    open(combinedFileName, 'wb').write(bigPdf)

def copyPages(frompdffile, tocanvas, withoutline=1):
    """copy all pages from pdf file into the Reportlab canvas."""
    copyPagesInMemory(nativeFromBytes(open_and_read(frompdffile)), tocanvas, withoutline)

_docCount = 0
def copyPagesInMemory(pdfFileContent, tocanvas, withoutline=1):
    """copy all pages from pdf content into the Reportlab canvas"""
    #need a doc-specific prefix
    global _docCount
    _docCount += 1
    p = PDFParseContext(pdfFileContent, asNative(_digester(pdfFileContent))+('%x' % _docCount))
    p.parse()
    (ind, catinfo) = p.catalog
    # this will add some extra gunk (non used catalog, for instance)
    c= p.compilation
    c.findAllReferences()
    catalog = c.getReference(catinfo)
    c.doTranslations(catinfo)
    # copy in all elements of the file
    docObject = tocanvas._doc
    from reportlab.pdfbase.pdfdoc import PDFZCompress, PDFStream
    for name in c.sequence:
        ob = c.objects[name]
        if isinstance(ob, PDFStream):
            # compress all streams
            ob.filters = [PDFZCompress]
        #print name, ob
        docObject.Reference(ob, name)
    # now add the pages
    c.populatePageList()
    pages = docObject.Pages
    pagesref = docObject.Reference(pages)
    for i,pageName in enumerate(c.pageList):
        pageObject = c.objects[pageName]
        pageObject["Parent"] = pagesref
        pages.addPage(pageObject)
    # add outline entries, if requested
    if withoutline:
        #print "catalog", catalog
        #print "catinfo", catinfo, ind
        catobj = c.resolve(catalog)
        #print "catobj", catobj
        if "Outlines" in catobj.dict:
            outlines = c.resolve(catobj.dict["Outlines"])
            #print "outlines", outlines
            #print "outlines dict", outlines.dict
            count = outlines.dict["Count"]
            if count:
                first = outlines.dict["First"]
                installOutlines(first, c, tocanvas,context=p)
                #stop
        for otherattribute in """Dests Threads AcroForm Names OpenActions URI
                ViewerPreferences PageLabels PageLayout JavaScript StructTreeRoot SpiderInfo""".split():
            if otherattribute in catobj.dict:
                value = catobj.dict[otherattribute]
                setattr(docObject.Catalog, otherattribute, value)
    #they might still be using the canvas, update its page number
    tocanvas._pageNumber = len(tocanvas._doc.Pages.pages) + 1
    return tocanvas

def _getDestOrADest(dict,context):
    Dest = dict.get('Dest',None)
    if "A" in dict:
        A = dict['A']
        if isinstance(A,pdfdoc.PDFObjectReference):
            if not context: return -1
            A = context.compilation.objects[A.name]
        if isinstance(A,pdfdoc.PDFDictionary):
            if ("S" in A.dict) and ('D' in A.dict):
                Dest = A.dict['D']
                return Dest
        return -1

def installOutlines(node, compilation, canvas, level=0, seen=None, context=None):
    c = compilation
    node = c.resolve(node)
    if seen is None: seen={}
    if node in seen:
        raise ValueError("infinite loop traversing outlines")
    seen[node] = 1
    dict = node.dict
    title = dict["Title"].s # get the title string
    #print "   "*level, title
    Dest = _getDestOrADest(dict,context)
    if Dest==-1:
        #print "warning: pagecatcher does not process actions in outlines"
        return
    if Dest:
        count = dict.get('Count',0)
        name = "destination"+ascii(id(Dest))
        canvas._destinations[name] = Dest
        #print "name, dest", (name, Dest)
        closed = 0
        if count<0: closed = 1
        canvas.addOutlineEntry(title, name, level, closed)
    # recursively do the children
    if "First" in dict:
        childnode = dict["First"]
        installOutlines(childnode, compilation, canvas, level+1, seen, context=context)
    # recursively do the syblings
    if "Next" in dict:
        siblingnode = dict["Next"]
        installOutlines(siblingnode, compilation, canvas, level, seen, context=context)

def pdfdocStr(s):
    try:
        s = s.decode(s.startswith(codecs.BOM_UTF16_BE) and 'utf16' or 'pdfdoc').encode('utf8')
    except:
        pass
    return s

# specialized info structures for recording product information
class PDFInfoL(pdfdoc.PDFInfo):
    """PDF documents can have basic information embedded, viewable from
    File | Document Info in Acrobat Reader.  If this is wrong, you get
    Postscript errors while printing, even though it does not print."""
    def __init__(self, oldinfo, user, number):
        pdfdoc.PDFInfo.__init__(self)
        self.title = pdfdocStr(oldinfo.title)
        self.author = pdfdocStr(oldinfo.author)
        self.subject = pdfdocStr(oldinfo.subject)
        self.keywords = pdfdocStr(oldinfo.keywords)
        self.producer = "(%s,%s) PageCatcher http://www.reportlab.com" % (ascii(user), ascii(number))

class PDFInfoU(pdfdoc.PDFInfo):
        title = "PageCatcher Evaluation Test Document"
        author = "PageCatcher Evaluator"
        producer = "Evaluation copy of PageCatcher http://www.reportlab.com"
        subject = ( "This document was created with an evaluation version of ReportLab PageCatcher.\n"
                    "Please contact ReportLab to obtain a license for production use.")

def storeForms(frompdffile, storagefile, pagenumbers=None, prefix="PageForms", all=None,
               BBoxes=0, fformname=None, password=""):
    "returns names of forms stored"
    if verbose: print("password is %s" % ascii(password))
    if pagenumbers is None:
        pagenumbers = [0] # default to first page
    if verbose: print("\n\n placing pages %s from %s into %s" % (pagenumbers, frompdffile, ascii(storagefile)))

    pdfContent = open_and_read(frompdffile)
    result = storeFormsInMemory(pdfContent, pagenumbers, prefix, all,
                BBoxes, #extractext=0,
                fformname=fformname, password=password)

    if BBoxes:
        #return value is a dict of bounding boxes;
        #preserve the old API
        (bboxInfoDict, pickledData) = result
        return bboxInfoDict
    else:
        (names, pickledData) = result
        if hasattr(storagefile, 'write'):
            storagefile.write(pickledData)
        else:
            if verbose: print("storing to %s" % ascii(storagefile))
            open(storagefile, 'wb').write(pickledData)
        return names

def storeFormsInMemory(pdfContent, pagenumbers=None, prefix="PageForms", all=None,
               BBoxes=0, extractText=0, fformname=None, password="",saveAnnots=False):
    """Stores Forms without touching disk.  You pass in the PDF file
    contents as a string.  It normally returns a 2-tuple with the list of form names
    (for backwards compatibility) and the content of the storage file, which
    is essentially a pickle of forms objects.

    If you set BBoxes=1, the first element is a dictionary mapping form names to page
    bounding boxes, so you can extract the page sizes, like this:
    {'thing0': [x0, y0, x1, y1],
     'page2': [x0, y0, x1, y1]}

    Extension in Aug 2006 to support ArtBox:  set BBoxes=2 and you get a dictionary
    mapping form names to a dictionary of box types to rects.
    {'thing0': {'MediaBox':[0,0,595,842], 'ArtBox':[20,15,300,475]}}
    
    """

    c, nameToObj = _storeFormsInDict(pdfContent, pagenumbers, prefix, all, fformname=fformname, password=password,saveAnnots=saveAnnots)
    pickledData = pickle.dumps(nameToObj, 1)
    formnames = nameToObj[None]
    result = (formnames, pickledData)
    if BBoxes == 1:
        # if the user requested bboxes then return dictionary name-->bbox
        bboxInfo = {}
        for formname in formnames:
            xobjname = pdfdoc.xObjectName(formname)
            form = nameToObj[ xobjname ]
            #print dir(firstform)
            dict = form.dictionary.dict
            #print dict
            bbox = dict["BBox"].sequence
            #print bbox
            bboxInfo[formname] = bbox
        result = (bboxInfo, pickledData)
    elif BBoxes == 2:
        # if the user requested bboxes=2 then return dictionary name --> box type --> bbox
        bboxInfo = {}
        for formname in formnames:
            xobjname = pdfdoc.xObjectName(formname)
            form = nameToObj[ xobjname ]
            #print dir(firstform)
            dict = form.dictionary.dict
            boxDict = {}
            for boxType in ['MediaBox','ArtBox','CropBox','TrimBox','BleedBox']:
                if boxType in dict:
                    boxDict[boxType] = dict[boxType].sequence
            bboxInfo[formname] = boxDict
        result = (bboxInfo, pickledData)
    elif extractText:
        result = c.plainText
    return result

def storeFormsInDict(pdfContent, pageNumbers=None, prefix="PageForms", all=None, fformname=None, password="",saveAnnots=False):
    """Stores Forms in a dict.  You pass in the PDF file
    contents as a string.  It returns a dict of form objects.
    To get form ie page names use the returnedValue[None]
    """
    return _storeFormsInDict(pdfContent, pagenumbers=pageNumbers, prefix=prefix, all=all,
                fformname=fformname, password=password,saveAnnots=saveAnnots)[0]

def _storeFormsInDict(pdfContent, pagenumbers=None, prefix="PageForms", all=None, fformname=None, password="", saveAnnots=False):
    # no security in this one... always store forms the same.
    p = PDFParseContext(pdfContent, prefix=prefix)
    p.parse()
    (ind, catinfo) = p.catalog
    c = p.compilation
    if p.encrypt:
        encryptinfo = p.resolve(p.encrypt)
        IDinfo = p.resolve(p.PDFID) # required for encryption
        IDarray = c.translateTree(IDinfo)
        #print dir(IDarray)
        #print ascii(IDarray.sequence[0])
        IDarray1 = unescapePDFCharacterSequence(IDarray.sequence[0])
        #print `IDarray1#`; burp
        #print encryptinfo
        EncryptionDict = c.translateTree(encryptinfo)
        #print EncryptionDict
        # next step: add decryptor info to the compilation structure.
        d = EncryptionDict.dict
##        for n in ("Filter", "R", "O", "U", "P", "V"):
##            print n, ascii(d[n])
##        for n in ("O", "U"):
##            print n, "value\t", d[n]
##            print n, "unescaped\t", ascii(unescapePDFString(d[n]))
        ### really should check to ensure standard encryption method
        OwnerKey = unescapePDFCharacterSequence(d["O"])
        UserKey = unescapePDFCharacterSequence(d["U"])
        #print len(d["U"]), ascii(d["U"])
        #print len(UserKey), ascii(UserKey)
        #stop
        Permissions = d["P"]
        c.encryptParameters(password, OwnerKey, UserKey, IDarray1, Permissions)
    c.sanitizePages(saveAnnots=saveAnnots)
    c.findAllReferences()
    catalog = c.getReference(catinfo)
    c.doTranslations(catinfo)
    c.populatePageList()
    return c, c.pagesAsForm(pagenumbers, prefix, all=all,fformname=fformname)

def restoreForms(storagefilename, canv, formnames=None,
                 allowDuplicates=0):
    if verbose: print("now restoring forms")
    content = open_and_read(storagefilename)
    return restoreFormsInMemory(content, canv, formnames=formnames, allowDuplicates=allowDuplicates)

def restoreFormsInMemory(content, canv, formnames=None, allowDuplicates=0):
    """Adds the named forms into the canvas and underlying PDF, with
    all the stuff they reference.
    If one attemtps to add forms twice, normally a ValueError is raised.
    If allowDuplicates == 1, it will test if any of the form names are present
    up front, and succeed silently (with few cycles) if they are.  this
    makes it safe to call it several times with the same form, which can
    be convenient.
    """
    if verbose: print("now restoring forms")
    return restoreFormsFromDict(pickle.loads(content), canv, formnames=formnames, allowDuplicates=allowDuplicates)

def hasFormNamed(canv, formName):
    """Returns whether a formName is already present in the canvas.

    Aids optimisations if repeatedly drawing, say, a PDF image"""
    return (pdfdoc.xObjectName(formName) in canv._doc.idToObject)

def getFormBoundingBox(canv, formName, boxType="MediaBox"):
    """Return bbox of the named form already present"""
    return canv._doc.getFormBBox(formName, boxType=boxType)

def _save_extra_info_and_clean_form(form,formName,nameToObj):
    fdict = form.dictionary.dict
    form._extra_pageCatcher_info = fxi = {}
    for k in _extra_info_keys:
        if k in fdict:
            v = fdict[k]
            del fdict[k]
            while isinstance(v,pdfdoc.PDFObjectReference):
                try:
                    v = nameToObj[v.name]
                except:
                    break
            if hasattr(v,'sequence'):
                v = v.sequence
            elif not isinstance(v,(int,float,str)):
                raise ValueError('Unknown extra info key type %r(%r) for %s in form %s' % (v,v0,k,formName))
            fxi[k] = v

def _nameToObjKeyFunc(k):
    if k is None:
        return (b'\x00',(-1000000,-1000000))
    elif isinstance(k,strTypes):
        return (asBytes(k),(1000000,1000000))
    elif isinstance(k,int):
        return (asBytes(str(k)),(0,0))
    else:
        return (asBytes(k[0]),)+k[1:]

def restoreFormsFromDict(nameToObj, canv, formnames=None, allowDuplicates=0):
    '''restores the forms from the dict nameToObj'''
    uses = nameToObj.get(0,{})
    # get the form names
    formNamesAll = nameToObj[None]
    if formnames is None:
        formnames = formNamesAll
    else:
        for fn in formnames:
            if fn not in formNamesAll:
                raise ValueError("no such form %s only found %s" % (
                    ascii(fn), formNamesAll))
    if verbose: print("formNames are %s" % ascii(formnames))

    doc = canv._doc
    #new mode allows us to safely call this repeated times, with
    #no effects - assuming form name is unique and reliable.
    if allowDuplicates:
        if verbose: print('up-front check forms are already present')
        discard = []
        for name in formnames:
            if pdfdoc.xObjectName(name) in doc.idToObject:
                discard.append(name)
                if verbose: print('%s already defined, not processing' % ascii(name))
        for d in discard:
            formnames.remove(d)
        if formnames == []:
            if verbose: print('no new forms, returning')
            return

    from rlextra.utils.ubold import W
    ANL = W.ANL
    ANP = W.ANP
    try:
        if verbose: W.pp('ACL')
        try:
            NS = {}
            rl_exec(W.AEXC, NS)
            License = NS['License']
        except:
            ANL = W.ANML
            if verbose>9: W.pp('ANML')
            raise
        if verbose>9:
            print("from %s" % ascii(License.__file__))
        number = str(License.number)
        user = License.user
        try:
            ANP = misc[0]
        except:
            pass
        if len(number)!=14:
            ANL = W.ANLL
            if verbose>9: W.pp('ANLL')
            raise ValueError
        key = License.key
        expire = License.expire
        misc = License.misc
        Sn = W.ASS+number+W.ASS+user+W.ASS+ascii(expire)+W.ASS+ascii(misc)
        code = md5(asBytes(Sn)).digest()
        from reportlab.lib.rl_accel import asciiBase85Encode
        code85 = asciiBase85Encode(code)
        if code85!=key:
            raise ValueError
        import time
        if expire is None:
            if number!=W.AMN or user!=W.AMU:
                #number is a date
                import rlextra
                from datetime import datetime
                def gd(n):
                    return datetime(int(n[0:4]),int(n[4:6]),int(n[6:8]))
                try:
                    d = rlextra.Version.split('.')[-1]
                    if len(d)==14: #daily
                        d = (gd(d) - gd(number)).days
                    else:
                        d = rlextra.__version__.split()[3].replace('-','')
                        d = (gd(d) - gd(number)).days
                except:
                    ANL = W.ANLC
                    if verbose>9: W.pp('ANLC')
                    raise ValueError
                else:
                    if d>366:
                        ANL = W.ANLM
                        if verbose>9: W.pp('ANLM')
                        raise ValueError
        elif expire<time.time():
            if verbose: W.pp('ALE',W.ALE % time.strftime('%Y%m%d',time.gmtime(expire)))
            raise ValueError
    except:
        ok = 0
        #import sys
        #print sys.exc_type, sys.exc_value
        doc.info = PDFInfoU() # mark file as unlicensed
        if verbose: W.pp('ALCF')
        # license failure: mess up the forms...
        # no comments in the output file
        pdfdoc.DoComments = 0
        #pdfdoc.DoComments = 1 # temp
        # encrypt all streams, make it hard to figure out how the encoding works.
        #doc.defaultStreamFilters = [pdfdoc.PDFZCompress]
        #doc.defaultStreamFilters = None #temp
        count = 0
        BBoxes = {}
        newNames = {}
        for formName in formnames:
            count += 1
            # create "wrapper" forms
            h = ascii(hash(formName))
            # make a unique new name that's hard to understand
            newName = "x%s%s%s" % (h[0:3], count, h[3:6])
            newNames[formName] = newName
            oldinternalname = pdfdoc.xObjectName(formName)
            # use the cryptic name for the form
            theform = nameToObj[oldinternalname]
            fdict = theform.dictionary.dict
            _save_extra_info_and_clean_form(theform,formName,nameToObj)
            thenewinternalname = pdfdoc.xObjectName(newName)
            #print "new", ascii(thenewinternalname)
            nameToObj[thenewinternalname] = theform
            uses[thenewinternalname] = uses.get(oldinternalname, [])
            # print thenewinternalname
            BBoxes[newName] = fdict["BBox"].sequence
            del nameToObj[oldinternalname]
##            # add marked form in place of the old one.
##            canv.beginForm(formName) # moved below
##            canv.doForm(newName)
##            canv.drawString(20, 20, "This page used an unlicensed copy of PageCatcher, http://www.reportlab.com.")
##            canv.endForm()
    else:
        for formName in formnames:
            _save_extra_info_and_clean_form(nameToObj[pdfdoc.xObjectName(formName)],formName,nameToObj)
        # license succeeded. put the license number into the info
        ok = 1
        doc.info = PDFInfoL(doc.info, user, number) # mark file as licensed
        if verbose: W.pp('ALO')
    # always compress streams
    if not ok or COMPACT:
        if verbose: print('ok=%s COMPACT=%s',(ok,COMPACT))
        doc.defaultStreamFilters = [pdfdoc.PDFZCompress]
    #print "\n\n\n Now installing objects in document"
    keys = list(nameToObj.keys())
    try:
        keys.sort(key=_nameToObjKeyFunc)
    except:
        print("keys = %s" % ascii(keys))
        rl_reraise()
    #print "uses", nameToObj[0]
    # install only objects used by these forms
    required = []
    if uses:
        for fn in formnames:
            efn = pdfdoc.xObjectName(fn)
            required.extend(uses[efn])
        installed = {}
        while required:
            require = required[-1]
            del required[-1]
            if not require in installed:
                if require in nameToObj:
                    ob = nameToObj[require]
                    doc.Reference(ob, require)
                    required.extend(uses[require])
                    #print 'installed', ob
                else:
                    #print 'CAN\'T FIND OBJECT', require
                    pass # if it can't be found it might be a "forbidden" object
                installed[require] = 1
    else:
        # if its using an old data file just install everything
        for name in keys:
            #print "installing", name
            ob = nameToObj[name]
            if name:
                #print name, ob
                doc.Reference(ob, name)
    # add form name translations
    if not ok:
        for (formName, newName) in list(newNames.items()):
            # add marked form in place of the old one.
            [lx, ly, ux, uy] = BBoxes[newName]
            canv.beginForm(formName, lx, ly, ux, uy)
            canv.doForm(newName)
            canv.drawString(20,20,ANL+ANP)
            canv.endForm()
            thenewinternalname = pdfdoc.xObjectName(newName)
            ob = nameToObj[thenewinternalname]
            doc.Reference(ob, thenewinternalname)
    else:
        for fn in formnames:
            efn = pdfdoc.xObjectName(fn)
            ob = nameToObj[efn]
            doc.Reference(ob, efn)
    return formnames

def fileName2Prefix(fileName,pfx='inclPDF'):
    return "%s:%s" % (pfx,asNative(_digester(fileName)))

def pdfImageFormDetails(canv,fileName,pageNumber,boxType='MediaBox'):
    '''
    determines form and dimensions for specified filename, pageNumber & boxtype
    '''
    prefix = fileName2Prefix(fileName)
    formName = prefix + str(pageNumber)

    if hasFormNamed(canv, formName):
        #print 'form %s already in canvas' % formName
        mediaX, mediaY, mediaX2, mediaY2 = getFormBoundingBox(canv, formName, boxType=boxType)
    else:
        #print 'storing form %s in canvas' % formName
        pdfContent = open_and_read(fileName, 'rb')

        sizeInfo, pageContent = storeFormsInMemory(
            pdfContent, pagenumbers=[pageNumber], prefix=prefix,
            BBoxes=2, password=PASSWORD
            )

        restoreFormsInMemory(pageContent,
                             canv,
                             formnames=[formName],
                             allowDuplicates=1)

        formName2 = list(sizeInfo.keys())[0]
        assert formName == formName2
        si = sizeInfo[formName]
        try:
            mediaX, mediaY, mediaX2, mediaY2 = si[boxType]
        except KeyError:
            #box we want might not be defined; they all default to MediaBox
            #according to the spec.  Use case: ten PDF images, half of which
            #have an ArtBox and half don't.
            mediaX, mediaY, mediaX2, mediaY2 = si["MediaBox"]
            
    mediaWidth = mediaX2 - mediaX
    mediaHeight = mediaY2 - mediaY
    return formName, (mediaX, mediaY, mediaWidth, mediaHeight)

def _getAdjustedDimensions(
        mediaX, mediaY, mediaWidth, mediaHeight,
        x=0, y=0, width=None, height=None,
        preserveAspectRatio=False,
        anchor='sw',
        ):
    adjustedX, adjustedY, adjustedWidth, adjustedHeight, scale = aspectRatioFix(
                                                            preserveAspectRatio,
                                                            anchor,
                                                            x,
                                                            y,
                                                            width,
                                                            height,
                                                            mediaWidth,
                                                            mediaHeight)
    adjustedX = adjustedX - (scale * mediaX)
    adjustedY = adjustedY - (scale * mediaY)

    return adjustedX,adjustedY,adjustedWidth,adjustedHeight

def _drawPdfForm(canv, formName,
                mediaX, mediaY, mediaWidth, mediaHeight,
                x=0, y=0, width=None, height=None,
                preserveAspectRatio=False,
                anchor='sw',
                showBoundary=False,
                ):

    adjustedX,adjustedY,adjustedWidth,adjustedHeight = _getAdjustedDimensions(
        mediaX, mediaY, mediaWidth, mediaHeight,
        x=x, y=y, width=width, height=height,
        preserveAspectRatio=preserveAspectRatio,
        anchor=anchor,
        )
    canv.saveState()
    canv.translate(adjustedX,adjustedY)
    canv.scale(adjustedWidth/float(mediaWidth),adjustedHeight/float(mediaHeight))
    canv.doForm(formName)
    canv.restoreState()

    if showBoundary:
        from reportlab.lib import colors
        canv.setLineWidth(0.5)

        #inner rect if preserveAspectRatio
        canv.setStrokeColor(colors.grey)
        canv.rect(adjustedX,
                  adjustedY,
                  adjustedWidth, adjustedHeight)

        if width and height:
            #requested outer rectangle
            canv.setStrokeColor(colors.black)
            canv.rect(x, y, width, height)

def drawPdfImage(fileName, canv, x=0, y=0, width=None, height=None,
                 preserveAspectRatio=False, pageNumber=0,
                 showBoundary=False,anchor='sw', boxType='MediaBox'):
    """Like a PDF version of canvas.drawImage...used by RML2PDF.

    This is designed to make it easy to place EPS artwork by distilling
    to PDF.
    - it will use the PDF file's MediaBox to determine the 'image' size, unless
      you explicitly request a different box such as the ArtBox in the boxType
      parameter  (drawings made in Illustrator may use this)

    - if x and y are not set, it presumes 0

    - if width and height are given, it will stretch and squeeze the image to
      fit in the given rectangle.

    - if width and height are NOT given, it will take the natural size as defined
      by the media box (or other box) of the PDF file

    - if width and height are given and preserveAspectRatio is set, it will
      shrink/grow the image to fit centered inside the defined box.  Note that
      it can only look at the bounding box of the EPS file and has no idea of
      the visual shape of whatever logos or designs that might contain!

    - showBoundary will draw a small line around the image box to help in
      seeing what's going on.

    """

    #we'll sneak a look inside to avoid even parsing the file
    #if the form already exists.  This is how we know the forms
    #will be named.
    #need to keep form names sane and free of slashes, protocols etc.

    formName, (mediaX, mediaY, mediaWidth, mediaHeight) = pdfImageFormDetails(canv,fileName,pageNumber,boxType)
        
    #now we have the form declared and know its size.  Ready to go...
    #ensure centred and shrunk in box if needed
    _drawPdfForm(canv, formName, mediaX, mediaY, mediaWidth, mediaHeight,
                x=x, y=y, width=width, height=height,
                preserveAspectRatio=preserveAspectRatio,
                showBoundary=showBoundary,anchor=anchor)

from reportlab.platypus.flowables import Image

class PDFImageFlowable(Image):
    """Flowable image which can accept PDF art too.
    
    This exists mainly so that the imageAndFlowables tag in RML can accept
    PDF files as images.  However it can also be used to insert a simple
    PDF image within a story.  More complex 'mini pages' with subtitles,
    alignment and so on are best handle with Figure tags."""
    def __init__(self, imageName, imageWidth, imageHeight):
        self.imageName = imageName
        self.imageWidth = imageWidth
        self.imageHeight = imageHeight
    def wrap(self, availWidth, availHeight):
        return (self.imageWidth, self.imageHeight)

    def _restrictSize(self,aW,aH):
        if self.imageWidth>aW+_FUZZ or self.imageHeight>aH+_FUZZ:
            factor = min(float(aW)/self.imageWidth,float(aH)/self.imageHeight)
            self.imageWidth *= factor
            self.imageHeight *= factor
        return self.imageWidth, self.imageHeight

    def draw(self):
        drawPdfImage(self.imageName, self.canv, width=self.imageWidth, height=self.imageHeight)

    

def page1(frompdffile, topdffile):
    "\n\n placing page 1 of", frompdffile, "in", topdffile
    storagefile = "dump.pickle"
    prefix = "myPrefix"
    names = storeForms(frompdffile, storagefile, [1], prefix, password=PASSWORD)
    canv = canvas.Canvas(topdffile)
    names = restoreForms(storagefile, canv, formnames=[prefix+"1"])
    # for testing, put a form in a form
    canv.beginForm("testform")
    canv.doForm(prefix+"1")
    canv.drawString(50, 500, "HELLO "*5)
    canv.endForm()
    canv.doForm("testform")
    print("\n\nnow writing %s" %ascii(topdffile))
    canv.save()
    print("page1 done")

def do_test_grids(storagefile, testfile):
    print("making test grid from %s into %s", (ascii(storagefile),ascii(testfile)))
    canv = canvas.Canvas(testfile, pageCompression=0)
    names = restoreForms(storagefile, canv)
    for name in names:
        print("form %s" % ascii(name))
        canv.doForm(name)
        canv.setFont("Helvetica", 6)
        from reportlab.lib.units import cm
        from reportlab.lib.colors import green, cyan
        canv.setStrokeColor(cyan)
        canv.setFillColor(green)
        lineend = cm*100 # too many cm, but what the...
        textoffset = 1.2*cm
        for count in range(100):
            y = count*cm
            canv.line(0, y, lineend, y)
            canv.drawString(textoffset, y, "y=%scm" % count)
        canv.rotate(90)
        for count in range(100):
            x = count*cm
            ty = -x
            canv.line(0, ty, lineend, ty)
            canv.drawString(textoffset, ty, "x=%scm" % count)
        canv.showPage()
    canv.save()
    print("wrote test file %s" % ascii(testfile))

def fourPage(storagefile, testfile, scalefactor = 0.5):
    print("placing forms from %s into %s four to a page", (ascii(storagefile),ascii(testfile)))
    from reportlab.pdfgen import canvas
    canv = canvas.Canvas(testfile)
    (width, height) = canv._pagesize
    names = restoreForms(storagefile, canv)
    put = sys.stdout.write
    while names:
        for (xoff, yoff) in [ (0,1), (1,1), (0,0), (1,0) ]:
            thisname = names[0]
            put(thisname+' ')
            canv.saveState()
            (x,y) = (xoff*width/2.0, yoff*height/2.0)
            canv.translate(x,y)
            canv.scale(scalefactor, scalefactor)
            canv.doForm(thisname)
            canv.restoreState()
            del names[0]
            if not names: break
        canv.showPage()
        print()
    canv.save()
    print("wrote %s" % ascii(testfile))

def cb_annotate(out, pdfin, cb=None):
    '''split out this callback version so others can get at the annotation themselves'''
    from rlextra.utils.proxyfile import MemFile
    storage = MemFile('')
    n = storeForms(pdfin, storage, [0], "annotate", password=PASSWORD)
    canv = canvas.Canvas(out)
    n = restoreForms(storage, canv)
    canv.doForm(n[0])
    if cb: cb(canv)
    canv.save()
    print("wrote %s" % ascii(out))

def annotate(out, pdfin, x,y, font, fontSize, text):
    def _cb(canv,x=x,y=y,font=font,fontSize=fontSize,text=text):
        canv.setFont(font, fontSize)
        canv.drawString(x,y,text)
    cb_annotate(out, pdfin, cb=_cb)

def test():
    import sys
    sstdout = sys.stdout
    fn = "dump.txt"
    #sys.stdout = open(fn, "w")
    sourcefile = "rfp_sample.pdf"
    page1(sourcefile, "onepage.pdf")
    #rewrite("translation1.pdf", "translation.pdf")
    sys.stdout = sstdout
    print("wrote %s"%fn)

#if __name__=="__main__": test()
#stop

script_usage = """
PAGECATCHER SCRIPT PREPROCESSOR USAGE:

In script mode the page catcher prepares the contents of one or more pages of
a PDF file for use in other PDF files.

Line mode usage:

% PageCatcher makeforms pdffile [-s storagefile] [-p prefix ] [-f formname]
-                               [--password password]
                                [--test pdftestfile] [--all] [pagenumber]*

  Capture the pages from pdffile and place them in storagefile for later use.
  If the test option is used then the captured pages are reimported and placed
  in the test file, overlayed with a centimeter grid.

  Note! Pagenumbers start at 0 (zero)
  (with no necessary relation to the pagenumber shown by a PDF viewer such as Acroread).
  If the pagenumbers are omitted the first page of the document will be made into
  a form.

  The "form names" for the forms derived from the pages will be prefix0 for the first page,
  prefix1 for the second, prefix2 for the third, and so forth.
  If the -f formname option is used then the first form only will be called formname.

  The password when provided should be the "read" (USER) encryption password
  or the pdffile.  Only provide the password option when the pdffile has be
  encrypted using the PDF 3.0 Standard Encryption method.

  For example

  % pageCatcher makeforms picture.pdf -s pic.data -p pict --test pictest.pdf 0 2

  extracts the first and third page from picture.pdf, archiving them in pic.data
  for later use, giving them the form names pict0 and pict2, respectively.
  The test file pictest.pdf will display the captured forms overlayed with a
  centimeter grid.

  If the storagefile is omitted it defaults to "storage.data".
  If the prefix is omitted it defaults to PF (for "page form").
  If the --all option is used then all pages of the document are captured.

USING THE DERIVED STORAGE FILES:

  To use the extracted forms later from a Python program proceed as follows:

    canvas = reportlab.pdfgen.Canvas(...)
    ...
    import pageCatcher
    # import pict0 and pict2 stored in pic.data by the command line above, put in canvas.
    storagefilename = "pic.data"
    formnames = pageCatcher.restoreForms(storageFileName, canvas)
    ...
    canvas.doForm(formnames[0]) # use the first form (as a backdrop for a page, for example).
    # or equivalently: canvas.doForm("pict0")
    ...
    canvas.save()

The restoreForms(filename, canvas) operation extracts the forms previously
stored in filename and installs them for later use in the pdfgen canvas object.

This is a licensed product.  You may use this module for evaluation purposes
but the forms will be visibly marked as "evaluation use of PageCatcher" and the
document will be interally annotated as a test document.

Licensed use of PageCatcher also results in the document being invisibly marked
with the license information.

DEMO SCRIPT MODES:

There are several built in demo modes:

 % pageCatcher.py help
     Print this message.

 % pageCatcher.py note [pdffile] [--password password]
     This mode places a text string over the first page of pdffile,
     storing the result in "annotated.pdf".  The font, size and
     string are read interactively from the console.

 % pageCatcher.py 4page pdffile [--scale scalefactor] [--password password]
                     [--output pdfoutputfilename] [-s storagefilename]
     This mode rewrites the pages of pdffile with 4 pages of the input
     on each page of the output ("save the trees" mode).

 % pageCatcher.py pages storagefile
 print names of the pages.
"""

#if __name__=="__main__":
def scriptInterp():
    try:
        import sys
        argv = list(sys.argv)
        all = None
        if "--all" in argv:
            argv.remove("--all")
            all = 1
        parameters = {"-s": "storage.data",
                      "--scale": 0.5,
                      "-p": "PF",
                      "-f": None,
                      "--test": None,
                      "--password": "",
                      "--output": None}
        for k in list(parameters.keys()):
            if k in argv:
                index = argv.index(k)
                del argv[index]
                value = argv[index]
                del argv[index]
                parameters[k] = value
        # demo modes use global variable to determine password
        global PASSWORD
        PASSWORD = parameters["--password"]
        #print 'pwd = <<', PASSWORD, '>>'
        known_modes = ["makeforms", "4page", "help", "note", "exec", "pages"]
        try:
            mode = argv[1]
        except:
            print("please provide a 'mode' argument")
            print("one of %s" % known_modes)
            return
        mode = mode.lower()
        if mode=="help":
            print(script_usage)
            return
        elif mode=="note":
            def sr(prompt):
                return input(prompt).strip()
            print("Annotate the first page of a file (interactive mode only)")
            try:
                pdffile = argv[2]
            except:
                pdffile = sr("Annotate pdf file: ")
            else:
                print("file=%s" %ascii(pdffile))
            try:
                x = float(sr("at x location (default 100): "))
            except:
                x = 100
            try:
                y = float(sr("at y location (default 400): "))
            except:
                y = 400
            font = sr("font [default=Helvetica]: ")
            if not font: font = "Helvetica"
            try:
                fontsize = float(sr("font size (default 20): "))
            except:
                fontsize = 20
            text = sr("Text: ")
            if not text: text = "Example Annotation"
            print("Attempting annotation...")
            annotate("annotated.pdf", pdffile, x, y, font, fontsize, text)
        elif mode=="4page":
            try:
                pdffile = argv[2]
            except:
                print("4Page requires a pdf file argument. Please provide filename")
                return
            storagefile = parameters["-s"]
            scale = parameters["--scale"]
            output = parameters["--output"]
            if output is None: output = "4page.pdf"
            print("Formatting %s as  %s 4 pages of input in each output page." % (ascii(pdffile),ascii(output)))
            print("Makeforms from %s to %s all pages" %(ascii(pdffile),ascii(storagefile)))
            names = storeForms(pdffile, storagefile, [], "fourpage", 1, password=PASSWORD)
            print("generating %s" % ascii(output))
            fourPage(storagefile, output, scale)
        elif mode=="makeforms":
            try:
                pdffile = argv[2]
            except:
                print("MakeForms requires a pdf file argument. Please provide filename")
                return
            storagefile = parameters["-s"]
            prefix = parameters["-p"]

            testfile = parameters["--test"]
            pagenumbers = argv[3:]
            if (pagenumbers and all):
                raise ValueError(
                      "Cannot use '--all' parameter together with specified page numbers %s" % ascii(pagenumbers))
            if not (pagenumbers or all):
                print("defaulting to first page")
                pagenumbers = [0]
            elif not all:
                pagenumbers = list(map(int, pagenumbers))
            else:
                pagenumbers = "all pages"
            print("Makeforms from %s to %s page numbers %s"%(ascii(pdffile),ascii(storagefile),pagenumbers))
            names = storeForms(pdffile, storagefile, pagenumbers, prefix, all, password=PASSWORD, fformname=parameters["-f"])
            if testfile is not None:
                do_test_grids(storagefile, testfile)
            print(); print("created forms: %s in %s"% (ascii(names),ascii(storagefile)))
        elif mode == 'exec':
            # execute a script
            try:
                scriptName = argv[2]
            except:
                print('Exec requires a python script name as the second argument.')
                return
            scriptname = argv[2]
            if not os.path.isfile(scriptname):
                print('Script "%s" not found!' % scriptname)
                return
            if __name__=='__main__':
                rl_exec(compile(open(scriptname).read(), scriptname, 'exec'),locals())
            else:
                oldname = globals().get('__name__', None)
                globals()['__name__'] = '__main__'
                rl_exec(compile(open(scriptname).read(), scriptname, 'exec'),locals())
                del globals()['__name__']
                if oldname: globals()['__name__'] = oldname
        elif mode == 'pages':
            try:
                P = _listPages(argv[2])
                s = ''
                for p in P:
                    if ' ' in p: p = '"'+p+'"'
                    s = '%s %s' % (s,p)
                print(s[1:])
            except:
                print('pages requires a storage file name as the second argument')
                raise
        else:
            print("unknown mode %s, known modes are %s" %(mode, known_modes))
    except:
        print(error_hint)
        print("ERROR %s %s" % (sys.exc_info()[0],sys.exc_info()[1]))
        if RERAISE:
            raise

def _listPagesInMemory(s):
   import pickle
   return pickle.loads(s)[None]

def _listPages(fn):
   return _listPagesInMemory(open_and_read(fn))

error_hint = """
PageCatcher has terminated in an error condition.

There are several common situations where this can occur:
  1) The input parameters may be invalid (the pdf file doesn't exist,
     or the pdf file doesn't contain the page you requested, etcetera).
  2) The PDF file may be encrypted with a non null password. If you have
     the owner password, save it without a password and try again.
  3) The page you requested contains active form elements, such as
     fillable fields or check buttons.  This version of pageCatcher does not
     correctly extract the graphics from these pages.
  4) The page you requested contains an array of streams one of which
     is encrypted using LZW encryption.  This version of pageCatcher does not
     correctly extract the graphics from these pages.
Often it is possible to "work around" problems such as 2 through 4 above
by making use of a "PDF printer emulator" such as Distiller.

If you have Distiller (or similar software) available, try this: open
the file in Acrobat or Acrobat Reader and "Print to PDF" using Distiller.
The resulting file will be visually identical to the original, and in all
cases we have observed it should be acceptible as a PageCatcher input.
"""
if __name__=="__main__": #NO RUNTESTS
    scriptInterp()
