#Copyright ReportLab Europe Ltd. 2000-2012
#see license.txt for license details
__version__=''' $Id: zipapp.py 41923 2012-11-09 15:18:55Z rgbecker $ '''
__all__=('ZipAppDef','ArgProcessor','ZipAppArchive','StandardDistros')
import sys, os, time, shutil, fnmatch
from .buildutils import  kill, chdir, path_join,\
                        path_dirname, isFile,\
                        isDir, fetchRL, svn_info, listAllFiles, listAllDirs,\
                        errorMsg, path_basename, findSVNURL,\
                        svn_update, stagingArea, rename, _externalRun
from glob import glob
from rlextra.utils.superglob import FileSet, SuperFileSet
from reportlab.lib.utils import get_rl_tempdir, rl_exec
from functools import reduce

class StandardDistros:
    pfb = 0
    def reportlabDistro(self,srcDir,full=0,buildDir=None,postcludes=[]):
        fs = FileSet(srcDir)
        fs.include('reportlab/**/*')
        fs.exclude('reportlab/**/*.pyc')
        fs.exclude('reportlab/**/*.pyo')
        fs.exclude('reportlab/**/CVS/*')
        fs.exclude('reportlab/**/.svn/**/*')
        fs.exclude('reportlab/**/.hg/**/*')
        fs.exclude('reportlab/**/PROJECT_CVS_TAG')
        fs.exclude('reportlab/**/filelist.txt')     #if we need these wel'll make our own I suppose
        sfs = SuperFileSet(fs)
        if not full:
            sfs.exclude('reportlab/graphics/samples/**/*')
            sfs.exclude('reportlab/demos/**/*')
            sfs.exclude('reportlab/docs/**/*')
            sfs.exclude('reportlab/test/**/*')
            sfs.exclude('reportlab/tools/**/*')
            sfs.exclude('reportlab/**/README.*')
            sfs.exclude('reportlab/**/README')
            sfs.exclude('reportlab/**/VERSION')
            sfs.exclude('reportlab/**/TODO')
            sfs.exclude('reportlab/**/changes')
            if not self.pfb:
                sfs.exclude('reportlab/fonts/**/*')
            sfs.exclude('reportlab/**/*.c')
            sfs.exclude('reportlab/**/*.h')
            sfs.exclude('reportlab/**/*.dsw')
            sfs.exclude('reportlab/**/*.dsp')
            sfs.exclude('reportlab/**/*.java')
            sfs.exclude('reportlab/**/Makefile.*')
        else:
            if buildDir:
                sfs = SuperFileSet(fs)
                fs = superglob.FileSet(buildDir)
                fs.include('reportlab/docs/*.pdf')
                sfs.append(fs)
        for d in postcludes: getattr(sfs,d[0])(d[1])
        return sfs

    def rlextraDistro(self,srcDir,full=0,postcludes=[]):
        fs = FileSet(srcDir)
        fs.include('rlextra/**/*')
        fs.exclude('rlextra/**/CVS/*')
        fs.exclude('rlextra/**/.svn/**/*')
        fs.exclude('rlextra/**/.hg/**/*')
        fs.exclude('rlextra/**/PROJECT_CVS_TAG')
        fs.exclude('rlextra/**/filelist.txt')       #if we need these wel'll make our own I suppose
        fs.exclude('rlextra/changeforms/**/*')
        fs.exclude('rlextra/delphi/**/*')
        fs.exclude('rlextra/designdocs/**/*')
        fs.exclude('rlextra/distro/**/*')
        fs.exclude('rlextra/dummydata/**/*')
        fs.exclude('rlextra/rlj/**/*')
        fs.exclude('rlextra/xmlforms/**/*')
        fs.exclude('rlextra/**/exe_stub*.*')
        fs.exclude('rlextra/**/do_freeze*.*')
        fs.exclude('rlextra/**/exemain*.*')
        fs.exclude('rlextra/rml2pdf/wordfilter/**/*')
        fs.exclude('rlextra/django/**/*')
        sfs = SuperFileSet(fs)
        if not full:
            sfs.exclude('rlextra/**/00readme.txt')
            sfs.exclude('rlextra/**/00README.txt')
            sfs.exclude('rlextra/**/*.bat')
            sfs.exclude('rlextra/bindings/**/*')
            sfs.exclude('rlextra/demos/**/*')
            sfs.exclude('rlextra/examples/**/*')
            sfs.exclude('rlextra/docs/**/*')
            sfs.exclude('rlextra/test/**/*')
            sfs.exclude('rlextra/tools/**/*')
            sfs.exclude('rlextra/utils/installer/**/*')
            sfs.exclude('rlextra/utils/tproj/**/*')
            sfs.exclude('rlextra/distro/**/*')
            sfs.exclude('rlextra/ers/examples/**/*')
            sfs.exclude('rlextra/pageCatcher/demos/**/*')
            sfs.exclude('rlextra/radxml/samples/**/*')
            sfs.exclude('rlextra/radxml/test/**/*')
            sfs.exclude('rlextra/rml2pdf/demos/**/*')
            sfs.exclude('rlextra/rml2pdf/doc/**/*')
            sfs.exclude('rlextra/rml2pdf/pythonfreezenotes.txt')
            sfs.exclude('rlextra/preppy/test/**/*')
            sfs.exclude('rlextra/preppy/flintstone*')
            sfs.exclude('rlextra/preppy/*.html')
            sfs.exclude('rlextra/preppy/*.prep')
            sfs.exclude('rlextra/**/README')
            sfs.exclude('rlextra/**/VERSION')
            sfs.exclude('rlextra/**/TODO')
            sfs.exclude('rlextra/**/changes')
            sfs.exclude('rlextra/graphics/barcode/**/*')
            sfs.exclude('rlextra/graphics/doc/**/*')
            sfs.exclude('rlextra/graphics/webcharts/**/*')
            sfs.exclude('rlextra/forms/test/**/*')
            sfs.exclude('rlextra/forms/examples/**/*')
            sfs.exclude('rlextra/forms/doc/**/*')
            sfs.exclude('rlextra/forms/javascript/**/*')
            sfs.exclude('rlextra/fundreporting/**/*')
            sfs.exclude('rlextra/java/**/*')
            sfs.exclude('rlextra/pageCatcher/**/*.bmp')
            sfs.exclude('rlextra/pageCatcher/**/*.pdf')
            sfs.exclude('rlextra/pageCatcher/**/*.html')
            sfs.exclude('rlextra/pageCatcher/**/*.rml')
            sfs.exclude('rlextra/pageCatcher/test/**/*')
            sfs.exclude('rlextra/rml2pdf/test/**/*')
            sfs.exclude('rlextra/security/**/*')
            sfs.exclude('rlextra/test/**/*')
        for d in postcludes: getattr(sfs,d[0])(d[1])
        return sfs

def _top_info(dir,R):
    for d in os.listdir(dir):
        D = path_join(dir,d)
        if isDir(D):
            if isDir(path_join(D,'.svn')):
                try:
                    R('URL=%(URL)s\n\tSVNVERSION=%(svnversion)s\n\tRevision=%(Revision)s\n\tLast Changed Rev=%(Last Changed Rev)s\n\tLast Changed Date=%(Last Changed Date)s\n\tLast Changed Author=%(Last Changed Author)s' % svn_info(D))
                except:
                    pass
            else:
                _top_info(D,R) #recur down one level

FCGI_TEMPLATES=dict(
        FCGI='''\
#!%(PYTHONEXE)s
#built at %(ZIPAPP_NOW)s
import os, sys
os.chdir(%(RAW)s'%(CGI2APPDIR)s')
sys.path.insert(0,os.path.dirname(os.getcwd()))
sys.path.insert(0,os.getcwd())
sys.path.insert(0,os.path.join(os.getcwd(),%(RAW)s"%(IZIPNAME)s"))
from %(appModuleName)s import %(appClassName)s
from rlextra.ers.apploggers import OptionalFileAppLogger
from rlextra.thirdparty import fcgi
%(appClassName)s.handleFCGI(
    OUTDIR=os.path.abspath(os.path.join(%(RAW)s'%(APP2OUTDIR)s','%(APPNAME)s')),
    HTTP_OUTDIR='%(HTTP_OUTDIR)s/%(APPNAME)s',
    logger = fcgi.Deferred(OptionalFileAppLogger,%(LOGDIR)s),%(app_keyword_arg_specs)s
    )
''',
    XFCGI='''\
#!%(PYTHONEXE)s
#built at %(ZIPAPP_NOW)s
import os, sys
os.chdir(%(RAW)s'%(CGI2APPDIR)s')
sys.path.insert(0,os.path.dirname(os.getcwd()))
sys.path.insert(0,os.getcwd())
sys.path.insert(0,os.path.join(os.getcwd(),%(RAW)s"%(IZIPNAME)s"))
from %(appModuleName)s import %(appClassName)s
from rlextra.ers.apploggers import OptionalFileAppLogger
from rlextra.thirdparty import fcgi
%(appClassName)s.earlyCheckFileTimeStamps=fcgi.CheckFileTimeStamps(files=%(FCGICHECKFILES)s,action=%(appClassName)s._fcgi_ecftscb)
%(appClassName)s.handleFCGI(
    OUTDIR=os.path.abspath(os.path.join(%(RAW)s'%(APP2OUTDIR)s','%(APPNAME)s')),
    HTTP_OUTDIR='%(HTTP_OUTDIR)s/%(APPNAME)s',
    logger = fcgi.Deferred(OptionalFileAppLogger,%(LOGDIR)s),%(app_keyword_arg_specs)s
    )
''',
    FCGIWSGI='''\
#!%(PYTHONEXE)s
#built at %(ZIPAPP_NOW)s
import os, sys
prog=os.path.abspath(sys.argv[0])
os.chdir(%(RAW)s'%(CGI2APPDIR)s')
sys.path.insert(0,os.path.dirname(os.getcwd()))
sys.path.insert(0,os.getcwd())
zip=os.path.join(os.getcwd(),%(RAW)s"%(IZIPNAME)s")
sys.path.insert(0,zip)
from %(appModuleName)s import %(appClassName)s
from rlextra.ers.apploggers import OptionalFileAppLogger
from rlextra.thirdparty import fcgiwsgi
%(appClassName)s.earlyCheckFileTimeStamps=fcgiwsgi.CheckFileTimeStamps(files=%(FCGICHECKFILES)s,action=%(appClassName)s._fcgi_ecftscb)
%(appClassName)s.handleFCGIWSGI(
    OUTDIR=os.path.abspath(os.path.join(%(RAW)s'%(APP2OUTDIR)s','%(APPNAME)s')),
    HTTP_OUTDIR='%(HTTP_OUTDIR)s/%(APPNAME)s',
    logger = fcgiwsgi.Deferred(OptionalFileAppLogger,%(LOGDIR)s),
    serverKwds=dict(
                    apache_content_length_fix=%(APACHE_CONTENT_LENGTH_FIX)s,
                    exitPPID=1,
                    shutdownCheck=fcgiwsgi.CheckFileTimeStamps(files=(zip,prog)),
                    ),%(app_keyword_arg_specs)s
    )
''',
    )

class ZipAppDef(StandardDistros):
    SELENIUM = None     #project relative folder of folder containing tests
    CGI_TEMPLATE = '''\
#!%(PYTHONEXE)s
#built at %(ZIPAPP_NOW)s
import os, sys
os.chdir(%(RAW)s'%(CGI2APPDIR)s')
sys.path.insert(0,os.path.dirname(os.getcwd()))
sys.path.insert(0,os.getcwd())
sys.path.insert(0,os.path.join(os.getcwd(),%(RAW)s"%(IZIPNAME)s"))
from %(appModuleName)s import %(appClassName)s
from rlextra.ers.apploggers import OptionalFileAppLogger
%(appClassName)s(
    OUTDIR=os.path.abspath(os.path.join(%(RAW)s'%(APP2OUTDIR)s','%(APPNAME)s')),
    HTTP_OUTDIR='%(HTTP_OUTDIR)s/%(APPNAME)s',
    logger = OptionalFileAppLogger(%(LOGDIR)s),%(app_keyword_arg_specs)s
    ).handleCgi()
'''

    pfb = 1
    def projectDistro(self,projectPath):
        raise ValueError('Abstract method called')

    def copy_project(self, project, stgdir, appdir):
        print('Copying %s files' % project)
        FS = self.projectDistro(path_join(stgdir,project))
        FS.copyFiles(appdir)

    def fetch_project(self, project, workDir, stgdir, target=None, branch=None):
        print('Fetching %s from repository%s' % (project,branch and  (' '+branch) or ''))
        svn_update(stgdir, findSVNURL(workDir,branch=branch), target=target or project)

    def fetch_selenium(self, project, stgdir, nosvn):
        seTests = self.SELENIUM
        if seTests:
            if isinstance(seTests,str): seTests = [seTests]
            L = []
            for p in seTests:
                p = path_join(stgdir, project, p,'tests')
                assert isDir(p),'Folder "%s" not found' % p
                L.append(p)
            seTests = L
            if not nosvn:
                svn_update(stgdir,'https://eiger/svn/private/trunk/thirdparty/selenium')
        self.seTests = seTests

    def copy_selenium(self, stgdir, appdir):
        seTests = self.seTests
        if seTests:
            fs = FileSet(stgdir)
            fs.include('selenium/**/*')
            for pat in ('**/.svn/**/*','docs/**/*',
                        ):
                fs.exclude('selenium/'+pat)
            seTarget = path_join(appdir,'selenium')
            fs.copyFiles(seTarget)
            for p in seTests:
                fs = FileSet(path_dirname(p))
                fs.include('tests/**/*')
                fs.exclude('**/.svn/**/*')
                fs.copyFiles(seTarget)

    def copy_vc(self, stgdir, appdir):
        print('Copying versionCheck')
        shutil.copy(os.path.join(stgdir,'rlinfra','dbtools','versionCheck.py'),appdir)

    def __init__(self,
                EXPLODE_PATTERNS = [],      #'/' separated super glob style patterns for relative files
                moduleName = '',            #the module name containing the WebApp class
                appClassName = '',      #the name of the WebApp class
                cgiName = '',           #the default name of the cgi script
                appName = '',               #default base name of the target folder
                EARLY_ACTIONS = [],
                STD_ACTIONS = [
                    'self.make_dirs({{RAW}}"{{APPDIR}}")',
                    'self.make_webapp_cgi("{{APPNAME}}",{{RAW}}"{{APPDIR}}","{{MODULENAME}}","{{APPCLASSNAME}}","{{CGINAME}}",{{RAW}}"{{CGIDIR}}",{{RAW}}"{{LOGSDIR}}",{{RAW}}"{{OUTDIR}}")''',
                    'self.copyto({{RAW}}"{{APPDIR}}",(self._fn,"{{IZIPNAME}}"))',
                    'self.explodefilesrelative({{RAW}}"{{APPDIR}}",P=meta_data["filesrelative"])',
                    'self.removefrominstalledzip(os.path.join({{RAW}}"{{APPDIR}}","{{IZIPNAME}}"),P=meta_data["filesrelative"])',
                    'self.checkdirs("{{LOGSDIR}}","{{OUTDIR}}")',
                    ],
                LATE_ACTIONS = [],
                EXTRA_VALUES = [],
                EXTRA_META_DATA = {},
                EXTRA_SUBSTITUTABLES = [],
                EXTRA_APP_KEYWORD_ARG_SPECS = [],
                **kw
                ):
        for k,v in list(locals().items()):
            if k not in ('self','kw'): setattr(self,k,v)
        self.__dict__.update(kw)

    def usage(self, msg='',exit=-1):
        import os
        msg = msg and [msg] or []
        msg += [
        'Use %s [options]' % os.path.basename(sys.argv[0]),
        '\toptions',
        '\t--help           give this help and exit',
        '\t--project=x      set version number',
        '\t--version=x      set version number',
        '\t--comment=x      zipapp comment',
        '\t--zipdir=x       where zip should be sent',
        '\t--appname=name   application name',
        '\t--appdir=dir     application output folder',
        '\t--clean          remove staging area',
        '\t--verbose[=int]  set verbosity',
        ]
        errorMsg('\n'.join(msg),exit)

    def _get_meta_data(self,**kw):
        meta_data = dict(
            substitutables = ('actions','app_keyword_arg_specs','filesrelative')+tuple(self.EXTRA_SUBSTITUTABLES),
            predefines = dict(APPNAME=self.appName,APPCLASSNAME=self.appClassName,MODULENAME=self.moduleName),
            values = [
                ('PYTHONEXE', 'sys.executable', 'python executable that should be used'),
                ('WWW_UID', '"nobody"', 'user name for permissions etc'),
                ('WWW_GID', '"nobody"', 'user name for permissions etc'),
                ('WEBSITE', 'os.path.join(os.getcwd())', 'here htdocs, cgi-bin etc live'),
                ('HTDOCSDIR', 'os.path.join({{RAW}}"{{WEBSITE}}","htdocs")', 'Where html documents live'),
                ('CGIDIR', 'os.path.join({{RAW}}"{{WEBSITE}}","cgi-bin")', 'Where cgi files live'),
                ('CGINAME', '"%s"' % self.cgiName,"Name of the CGI script"),
                ('APPDIR', 'os.path.join({{RAW}}"{{WEBSITE}}","{{APPNAME}}")','Application folder'),
                ('LOGSDIR', 'os.path.join({{RAW}}"{{WEBSITE}}","logs")','Where logging folders live (none for no logging)\n'),
                ('OUTDIR', 'os.path.join({{RAW}}"{{WEBSITE}}","pdfout")','Where output documents may be written'),
                ('HTTP_OUTDIR', '"/pdfout"', 'HTTP name of the output folder'),
                ('APP2OUTDIR', 'self.relativeLocation({{RAW}}"{{APPDIR}}",{{RAW}}"{{OUTDIR}}")', ''),
                ('CGI2APPDIR', 'self.relativeLocation({{RAW}}"{{CGIDIR}}",{{RAW}}"{{APPDIR}}")', ''),
                ]+self.EXTRA_VALUES,
            filesrelative = self.EXPLODE_PATTERNS, #files to be expanded relative (may include glob patterns)
            actions = self.EARLY_ACTIONS + self.STD_ACTIONS + self.LATE_ACTIONS,
            app_keyword_arg_specs = self.EXTRA_APP_KEYWORD_ARG_SPECS,
            comment=getattr(self,'_comment',''),
            )
        meta_data.update(self.EXTRA_META_DATA)

        if getattr(self,'CGI',True):
            meta_data['CGI_TEMPLATE'] = self.CGI_TEMPLATE
        meta_data.update(kw)
        return meta_data

    def post_copy(self):
        pass

    def build(self):
        AP = ArgProcessor(sys.argv)
        help = AP.getFlag('--help')
        aproject = AP.getFlag('--project',1,getattr(self,'project',None))
        aversion = AP.getFlag('--version',1,None)
        self.verbose = int(AP.getFlag('--verbose',1,0))
        self._comment = AP.getFlag('--comment',1,'')
        vc = int(AP.getFlag('--vc',1,getattr(self,'VERSIONCHECK',0)))
        zipDir = AP.getFlag('--zipdir',1,None)
        repos = self.repos
        clean = AP.getFlag('--clean')
        self.appName = AP.getFlag('--appname',1,self.appName)
        startDir = path_dirname(sys.argv[0]) or os.getcwd()
        workDir = chdir(startDir)
        if not (aproject and aversion):
            project, version = open('version.txt','r').readlines()[0].split()[0:2]
        if aversion: version = aversion
        if aproject: project = aproject
        timeStamp = time.strftime('%Y%m%d%H%M%S',time.gmtime())
        project_version = '%s-%s-%s' % (project,version,timeStamp)
        appdir = self.appdir = AP.getFlag('--appdir',1,
                        getattr(self,'appdir',
                                get_rl_tempdir(os.path.join(project+'_buildarea',project))))
        I = AP.invalidFlags(())
        if I: self.usage('Invalid options %s' % (', '.join(I)))
        if help: self.usage(exit=0)
        chdir(workDir)
        def report(*args):
            x = ' '.join(args)
            sys.stdout.write(x)
            if not x.endswith('\n'):
                sys.stdout.write('\n')
        report('Removing build dir', appdir)
        kill(appdir)

        report('Making build dir', appdir)
        os.makedirs(appdir)
        self.version = version

        stgdir = stagingArea(self.stgdir)

        if clean and (isDir(stgdir) or isFile(stgdir)):
            report('Removing staging area', stgdir)
            kill(stgdir)

        for repo in repos:
            report('##### processing repo %r' % repo.url)
            repo.report = report 
            repo.fetch(stgdir)
            repo.copy(stgdir,appdir, pfb=repo.url.strip('/').endswith('/reportlab') and self.pfb)

        self.post_copy()

        #obtain revisions etc for toplevel staged
        _ = []
        _top_info(stgdir,_.append)

        #store marshalled metadata
        f = open(path_join(appdir,'meta_data.mar'),'wb')
        import marshal
        marshal.dump(self._get_meta_data(project_version=project_version,top_info='\n\n'.join(_)),f)
        f.close()

        if vc: self.copy_vc(stgdir,appdir)

        L = getattr(self,'LICENSES',{})
        L.update(getattr(self,'LICENCES',{}))

        for k,v in list(L.items()):
            f = open(os.path.join(appdir,k+'License.py'),'w')
            f.write("'''%s License'''\n" % k)
            for k in ('number','key','user','expire','misc'):
                f.write('%s = %s\n' % (k,repr(v[k])))
            f.close()

        try:
            self.make_zip(project,path_join(zipDir or workDir,'%s.zip' % project_version),stgdir,appdir)
        except:
            print('!!!!! error in make_zip\n!!!!! zipDir=%r\n!!!!! workDir=%r\nproject_version=%r\n!!!!! stgdir=%r\n!!!!!appdir=%r' % (zipDir,workDir,project_version,stgdir,appdir))
            raise

    def make_zip(self,project,zip,stgdir,appdir):
        #we do the compile and build using the staging dir as environment
        ifn = os.path.join(stgdir,'_externalRun.tmp')
        kill(ifn)
        cmd = sys.executable + ''' -c"from rlextra.utils.buildutils import _compileAppX;_compileAppX(r'%s','%s',r'%s',r'%s')"''' % (ifn,project,appdir,zip)
        try:
            _externalRun(ifn,cmd,None,verbose=1,errMsg='Errors during app compilation')
        except:
            print('!!!!! ifn=%s' % ifn)
            print('!!!!! cmd=%s' % cmd)
            print('!!!!! stgdir=%s' % stgdir)
            raise

def _replace(obj,s,r):
    try:
        return obj.replace(s,r)
    except AttributeError:
        return obj

def _substitute_values(s,v,known,V):
    if type(v)!=type(''): v = str(v)
    known[s] = v
    r = '{{%s}}' % s
    for i in range(len(V)):
        try:
            t0,t1,t2 = list(V[i])
            t1 = _replace(t1,r,v)
            t2 = _replace(t2,r,v)
            V[i] = (t0,t1,t2)
        except:
            print('ERROR in substitution of r=%s into V[%d]=%r\nV=[%s]' % (str(r),i,V[i],',\n'.join(map(str,V))))
            import traceback
            traceback.print_exc()

from .manifester import FilenameClassifier
_classifier_mode = FilenameClassifier().mode
del FilenameClassifier

class ArgProcessor:
    def __init__(self,argv,startAt=1):
        self.argv = argv[startAt:]
    def getFlag(self,name,nextVal=0,default=None):
        r = name in self.argv
        if nextVal and not r:
            name += '='
            for x in self.argv:
                if x.find(name)==0:
                    name = x
                    r = '='.join(name.split('=')[1:])
                    nextVal = 0
                    break
        if r:
            x = self.argv.index(name)
            del self.argv[x]
            if nextVal:
                try:
                    r = self.argv[x]
                    del self.argv[x]
                except:
                    r = (nextVal>>1)&1
        else:
            r = default
        return r

    def invalidFlags(self,allowed):
        I = []  #invalid flags
        for a in self.argv:
            if a[0]=='-':
                if a[:2]=='--' and a.find('=')>0: a = a.split('=')[0]
            if a not in allowed: I.append(a)
        return I

def _realpath(x):
    try:
        x = os.path.realpath(x)
    except:
        pass
    return os.path.abspath(os.path.normcase(os.path.normpath(x)))

def _relativeLocation(basedir,target):
    if not target: return ''
    basedir = _realpath(basedir)
    target = _realpath(target)
    BL = basedir.split(os.sep)
    TL = target.split(os.sep)
    i = 0
    nBL = len(BL)
    nTL = len(TL)
    n = min(nBL,nTL)
    while i<n and BL[i]==TL[i]: i += 1
    return os.sep.join((nBL-i)*['..']+TL[i:])

def _rawChar():
    return os.sep=='\\' and 'r' or ''

def _targetnorm(target,dn):
    if target and not os.path.isabs(target):
        target = os.path.normpath(os.path.join(dn,target))
    return target

def _src2code(src,target_pyc):
    #code ripped from imputil
    import imp, struct, marshal, builtins
    code = builtins.compile(src,target_pyc,'exec')
    f = []
    f.append(imp.get_magic())
    f.append(struct.pack('<I', int(time.time())))
    f.append(marshal.dumps(code))
    return ''.join(f)

def _make_keyword_arg_specs(*objs):
    K = {}
    for obj in objs:
        for x in obj:
            try:
                k,v = x.split('=',1)
                K[k] = v
            except:
                raise ValueError('Bad app_keyword_arg spec """%s"""' % x)
    return ["%s=%s" % k for k in list(K.items())]

class _ZAA_Importer:
    '''An importer object intended to allow our dynamic things to be imported during the install'''
    def __init__(self):
        self.known = {}

    def add(self, n, c):
        n = os.path.splitext(n)[0].replace(os.sep,'.').replace('/','.')
        self.known[n] = c

    def find_module(self,fullname,path=None):
        try:
            self._get_code(fullname)
            return self
        except:
            return None

    def _get_code(self,fullname):
        try:
            return self.known[fullname]
        except KeyError:
            raise ImportError('Unknown module %s'%fullname)

    def load_module(self, fullname):
        import imp, marshal
        code = self._get_code(fullname)
        if code[0:4]!=imp.get_magic():
            raise ValueError('Bad magic number in zip entry "%s"' % script)
        code = marshal.loads(code[8:])
        mod = imp.new_module(fullname)
        sys.modules[fullname] = mod
        mod.__file__ = "<%s>" % self.__class__.__name__
        mod.__loader__ = self
        mod.__path__ = []
        rl_exec(code, mod.__dict__)
        return mod

class ZipAppArchive:
    '''Encapsulate a zip archive for installable package uses'''
    def __init__(self,fileName):
        import zipimport
        self._ifn = self._fn = os.path.normcase(os.path.normpath(os.path.abspath(fileName)))
        zi = self._zi = zipimport.zipimporter(self._fn)
        self._files = list(zi._files.keys())
        self._unremoved = []
        self._unremovedD = []
        self._retained = []
        self._precious = []
        self._logfile = None
        self._dynamic_entries = []
        self._importer = _ZAA_Importer()

    def get_data(self,name):
        '''this should be a relative name'''
        return self._zi.get_data(os.path.join(self._fn,name))

    def _add_unremoved(self,d):
        unr = self._unremoved
        d = os.path.abspath(d)
        if isFile(d):
            unr.append(d)
        else:
            for _ in map(os.path.normcase,listAllFiles(d,rel=0)):
                if _ not in unr:
                    unr.append(_)
            und = self._unremovedD
            for _ in map(os.path.normcase,listAllDirs(d,rel=0)):
                if _ not in und:
                    und.append(_)

    def make_dirs(self, *dirs):
        precious = self._precious
        for d in dirs:
            if d==os.sep: raise ValueError("Not allowed to use file %s" % d)
            if isDir(d) or isFile(d):
                nv = self._overwrite and 'y' or self.yesnoquery('Path %s exists! OK to remove [NO]? ' % d )
                if nv.lower() in ['y','yes']:
                    if precious:
                        self._add_unremoved(d)
                    else:
                        kill(d)
                        os.mkdir(d)
                else:
                    if isDir(d):
                        nv = self.yesnoquery('Continue with installation [N]? ')
                    else:
                        nv = 'no'
                    if nv.lower() in ['y','yes']:
                        self._add_unremoved(d)
                        return
                    sys.stderr.write('Cannot continue please back up "%s" and retry\n' % d)
                    os._exit(0)
            else:
                os.makedirs(d)

    def ensure_dirs(self,*dirs):
        for d in dirs:
            if d==os.sep: raise ValueError("Not allowed to use file %s" % d)
            try:
                os.makedirs(d)
            except:
                if not isDir(d):
                    raise ValueError('Cannot ensure %r' % d)

    def copyto(self,dst,*files):
        if not isDir(dst): self.make_dirs(dst)
        for f in files:
            if type(f) is tuple:
                shutil.copy(f[0],dst)
                fn = path_join(dst,f[1])
                rename(path_join(dst,path_basename(f[0])),fn)
                if f[0]==self._fn:
                    self._ifn = fn
            else:
                shutil.copy(f,dst)
                fn = f
            try:
                self._unremoved.remove(os.path.normcase(fn))
            except:
                pass
            self._log('Installed File      "%s"' % fn)

    def explodefilesrelative(self,dst, P=[], N=[]):
        '''copy things glob matching P[i] to dst''' 
        from operator import add as operator_add
        if not isDir(dst): self.make_dirs(dst)
        filt = lambda x,_files=self._files: fnmatch.filter(_files,x)
        F = [x.replace(os.sep,'/') for x in reduce(operator_add,[filt(p) for p in N],[])]
        F = [f for f in [x.replace(os.sep,'/') for x in reduce(operator_add,[filt(p) for p in P],[])] if f not in F]
        for f in F:
            m = _classifier_mode(f)
            d = self.get_data(f)
            if m=='t': d = d.replace('\r\n','\n').replace('\r','\n')
            self.writefile(os.path.join(dst,f), m, d)

    def removefrominstalledzip(self,installedzipfn,P=[],N=[]):
        '''remove things that match P, but not if they match N'''
        from zipfile import ZipFile
        from operator import add as operator_add
        oldname=installedzipfn+'.old'
        rename(installedzipfn,oldname)
        old = ZipFile(oldname)
        P = P+['selenium/*']
        filt = lambda x,_files=self._files: fnmatch.filter(_files,x)
        F = [x.replace(os.sep,'/') for x in reduce(operator_add,[filt(p) for p in N],[])]
        F = [f for f in [x.replace(os.sep,'/') for x in reduce(operator_add,[filt(p) for p in P],[])] if f not in F]
        new = ZipFile(installedzipfn,'w')
        while self._dynamic_entries:
            name, entry = self._dynamic_entries.pop(0)
            new.writestr(name,entry)
        for zi in old.infolist():
            fn = zi.filename
            if fn not in F:
                new.writestr(zi,old.read(fn))
        old.close()
        new.close()
        kill(oldname)

    def writefile(self,dst,mode,data):
        try:
            os.makedirs(os.path.dirname(dst))
            self._log('Installed Directory "%s"' % fn)
        except:
            pass
        try:
            self._unremoved.remove(os.path.normcase(os.path.abspath(dst)))
        except:
            pass
        if isFile(dst):
            f = open(dst,'r'+ (('b' in mode) and 'b' or 't'))
            odata = f.read()
            f.close()
            if odata==data:
                self._retained.append(dst)
                return
            nv = self._overwrite and 'y' or self.yesnoquery('File %s already exists OK to replace [NO]? ' % dst)
            if nv.lower() not in ['y','yes']:
                nv = self.yesnoquery('Continue with installation [NO]? ')
                if nv.lower() in ['y', 'yes']: return
                os._exit(0)
        if 'w' not in mode: mode = 'w'+mode
        f = open(dst,mode)
        f.write(data)
        f.close()
        self._log('Installed File      "%s"' % dst)

    def make_webapp_cgi(self,appName,appDir,appModuleName,appClassName,cgiName,cgiDir,logsDir,outDir,
            extra_app_keyword_arg_specs=[]):
        '''this will only be called when we have all values setup in self.'''
        ZIPAPP_NOW = time.strftime('%Y%m%dU%H%M%S',time.gmtime())
        known = self._known.copy()
        if logsDir.lower()=='none':
            LOGDIR = 'None'
        else:
            LOGDIR = '''os.path.abspath(%s'%s')''' % (self._known['RAW'],self.relativeLocation(appDir,logsDir))

        app_keyword_arg_specs = _make_keyword_arg_specs(
                    self._meta_data.get('app_keyword_arg_specs',[]),
                    extra_app_keyword_arg_specs)
        app_keyword_arg_specs = app_keyword_arg_specs and ('\n\t'+(',\n\t'.join(app_keyword_arg_specs))+',') or ''
        known.update(locals())
        #now do substitution
        meta_data = self._meta_data
        if 'CGI_TEMPLATE' in meta_data:
            text = meta_data['CGI_TEMPLATE'] % known
            self.make_script(text,cgiDir,cgiName)
        FCGI = known.get('FCGI','')
        if FCGI:
            tmpl = FCGI_TEMPLATES.get(FCGI,None)
            if tmpl:
                text = tmpl % known
                self.make_script(text,cgiDir,os.path.splitext(cgiName)[0]+'.fcgi')

    def make_script(self,text,scriptDir,scriptName):
        print("make_script: %s/%s" % (scriptDir,scriptName))
        scriptPath = os.path.join(scriptDir,scriptName)
        self.writefile(scriptPath,"t",text)
        self.chmod(0o755,scriptPath)

    def make_config(self,pyc,**kw):
        code = []
        for k,v in list(kw.items()):
            code.append('%s = %s' % (k,repr(v)))
        code = _src2code('\n'.join(code)+'\n',pyc)
        self._dynamic_entries.append((pyc, code))
        self._importer.add(pyc,code)

    def make_link(self,src,tgt):
        kill(tgt)
        if sys.platform=='win32':
            shutil.copy(src,tgt)
            self._log('copy(%r,%r)' % (src,tgt))
        else:
            src = self.relativeLocation(path_dirname(tgt),src)
            os.symlink(src,tgt)
            self._log('symlink(%r,%r)' % (src,tgt))

    def platformValue(**kwds):
        p = sys.platform
        try:
            default = kwds.pop('default')
            return kwds.get(p,default)
        except KeyError:
            return kwds[p]
    platformValue=staticmethod(platformValue)

    def yesnoquery(self,prompt):
        yn = getattr(self,'_yesno',0)
        if not yn: return input(prompt)
        return yn<0 and 'n' or 'y'

    def showMessage(self,msg, dolog=1):
        print(msg)
        if dolog: self._log(msg.strip())

    def runScriptFromFS(self,script,question='',workDir=None,argv=None):
        '''ask whether an existing script should be executed at install time
            if the answer is yes then execute the script
        '''
        if question and self.yesnoquery(question+'[NO]?').lower() not in ('y','yes'): return
        omods = sys.modules.copy()
        cwd = os.getcwd()
        oargv = sys.argv[:]
        import __main__
        mD = __main__.__dict__
        omD = mD.copy()
        mD.clear()
        mD.update({'__name__': '__main__', '__builtins__':__builtins__,'__doc__':'','__file__': script})
        try:
            script = os.path.abspath(script)
            import imputil
            if not imputil._os_stat: imputil._os_bootstrap()
            if script[-4:] in ('.pyc','.pyo'):
                py = script[:-4]+'.py'
                if isFile(py):
                    code = imputil.py_suffix_importer(py,os.stat(py),None)[1]
                else:
                    import imp, marshal
                    f = open(script, 'rb')
                    assert f.read(4)==imp.get_magic(), 'Bad magic number in file "%s"' % script
                    t = f.read(4)   #time stamp bytes
                    code = marshal.load(f)
                    f.close()
            elif script.endswith('.py'):
                code = imputil._compile(script,imputil._timestamp(script))
            else:
                raise ValueError('Don\'t know how to execute script "%s"' % script)
            if workDir: os.chdir(workDir)
            if self._fn not in sys.path: sys.path.insert(0,self._fn)
            sys.argv[0] = script
            if self._importer.known: sys.meta_path.insert(0,self._importer)
            rl_exec(code, mD)
            if self._importer.known:
                try:
                    sys.meta_path.remove(self._importer)
                except:
                    pass
        finally:
            if workDir: os.chdir(cwd)
            mD.clear()
            mD.update(omD)
            from sets import Set
            for k in Set(list(sys.modules.keys()))-Set(list(omods.keys())):
                del sys.modules[k]
            for k, v in list(omods.items()):
                sys.modules[k] = v
            sys.argv = oargv
            del omods

    def runScriptFromZip(self,script,question='',workDir=None):
        '''run a script from the zip file'''
        if question and self.yesnoquery(question+'[NO]?').lower() not in ('y','yes'): return
        omods = sys.modules.copy()
        cwd = os.getcwd()
        import __main__
        mD = __main__.__dict__
        omD = mD.copy()
        mD.clear()
        mD.update({'__name__': '__main__', '__builtins__':__builtins__,'__doc__':'','__file__': script})
        try:
            oargv = sys.argv[:]
            a0 = sys.argv[0] = os.path.join(self._fn,script)
            if workDir: os.chdir(workDir)
            code = self.get_data(script)
            import imp, marshal, builtins
            if script[:-3]=='.py':
                try:
                    code = builtins.compile(code,script,'exec')
                except Exception as e:
                    args = list(e.args)
                    if args: args[0] = 'Failed to compile zip entry "%s"\n%s' % (script,args[0])
                    raise e.__class__(*args).with_traceback(sys.exc_info()[3])
            else:
                if code[0:4]!=imp.get_magic():
                    raise ValueError('Bad magic number in zip entry "%s"' % script)
                code = marshal.loads(code[8:])
            if self._fn not in sys.path: sys.path.insert(0,self._fn)
            rl_exec(code, mD)
        finally:
            if workDir: os.chdir(cwd)
            mD.clear()
            mD.update(omD)
            from sets import Set
            for k in Set(list(sys.modules.keys()))-Set(list(omods.keys())):
                del sys.modules[k]
            for k, v in list(omods.items()):
                sys.modules[k] = v
            sys.argv = oargv
            del omods

    def makeStub(self,entryName,target_pyc=None,target_py=None,zdn=None,ifn=None,preInit=[],postInit=[]):
        if not ifn: ifn = self._ifn
        zdn = zdn or os.path.dirname(ifn)
        target_py = _targetnorm(target_py,zdn)
        target_pyc = _targetnorm(target_pyc,zdn)
        src = '\n'.join([
            '''#!%s -u''' % sys.executable,
            '''import sys, os''',
                    ]+preInit+[
            '''zfn = os.path.join(os.path.abspath(os.path.dirname(sys.argv[0]) or os.getcwd()),%s"%s")''' % (_rawChar(),self.relativeLocation(os.path.dirname(target_py or target_pyc),os.path.join(zdn,os.path.basename(ifn)))),
            '''sys.path.insert(0,zfn)''',
                    ]+postInit+[
            '''from rlextra.utils.zipapp import ZipAppArchive''',
            '''ZipAppArchive(zfn).runScriptFromZip("%s")''' % entryName,
            ])+'\n'

        if target_py:
            self.writefile(target_py,'t',src)
            self.chmod(0o755,target_py)

        if target_pyc:
            self._log('Made compiled stub script "%s"' % target_pyc)
            self.writefile(target_pyc,'b',_src2code(src,target_pyc))
            self.chmod(0o755,target_pyc)
            if not target_py: self._log('%s\n%s\n%s' % (40*'#',src,40*'#'))

    def relativeLocation(self,basedir,target):
        return _relativeLocation(basedir,target)

    def _log(self,msg):
        if self._logfile: self._logfile.write(msg+'\n')

    def _cp_get(self,s,v):
        known = {}
        for k,x in list(self._known.items()):
            known[k.lower()] = x
        for sect in (self._meta_data['project_version'].split('-')[0],'zipapp'):
            try:
                return self._cp.get(sect,s.lower(),0,known)
            except:
                pass
        return v

    def _get_meta_data(self):
        import marshal
        return marshal.loads(self.get_data('meta_data.mar'))

    def install(self):
        self._logfilename = path_join(get_rl_tempdir('installs'),os.path.splitext(path_basename(self._fn))[0]+'.install.log')
        self._logfile = open(self._logfilename,'w')
        from rlextra.utils.cgisupport import TimeStamp
        self._log('Install carried out at %s GMT' % TimeStamp().timeStamp)
        self.showMessage("*NB* absolute path names should be entered, but will be relativized\nbefore use, so all can be moved bodily together!\n",dolog=0)
        AP = ArgProcessor(sys.argv)
        yesno = AP.getFlag('--yes')
        if AP.getFlag('--no'): yesno = -1
        self._yesno = yesno
        config = AP.getFlag('--config',1,None)
        if not config:
            HOME = os.environ.get('HOME',None)
            if HOME: config = os.path.join(HOME,'.zipapp.rc')
        self._meta_data = meta_data = self._get_meta_data()
        predefines = meta_data['predefines']
        appname = predefines['APPNAME']
        predefines['APPNAME'] = AP.getFlag('--appname',1,predefines['APPNAME'])
        if config and isFile(config):
            from configparser import SafeConfigParser
            cp = SafeConfigParser(predefines)
            cp.read(config)
        else:
            cp = None
        self._cp = cp
        self._known = known = {}
        _ = ('y','yes','1','true')
        self._overwrite = AP.getFlag('--overwrite') or self._cp_get('overwrite','n').lower() in _
        forceInstall = AP.getFlag('--overwrite') or self._cp_get('overwrite','n').lower() in _
        goodConfig = AP.getFlag('--goodconfig') or self._cp_get('goodconfig','n').lower() in _

        V =  meta_data['values']
        zip = os.path.basename(self._fn)
        izip = os.path.splitext(zip)
        izip = '-'.join(izip[0].split('-')[:-2])+izip[1]
        _substitute_values('ZIPNAME',zip,known,V)
        _substitute_values('IZIPNAME',izip,known,V)
        _substitute_values('RAW',_rawChar(),known,V)
        for k,v in list(predefines.items()):
            _substitute_values(k,v,known,V)
        NV = []
        for s,v,desc in V:
            nv = AP.getFlag('--%s' % s.lower(),1,None)
            if nv is not None:
                _substitute_values(s,nv,known,V)
                continue
            else:
                nv = self._cp_get(s,v)
                if nv is not v:
                    if goodConfig:
                        _substitute_values(s,nv,known,V)
                        continue
                    else:
                        v = '"%s"' % nv
            NV.append((s,v,desc))
        V = NV

        n = len(V)*10
        while V and n>0:
            s,v,desc = V.pop(0)
            n -= 1
            if s in list(known.keys()): continue
            if isinstance(v,str):
                if '{{' in v:
                    V.append((s,v,desc))    #move to end
                    continue
                v = eval(v)
            if desc:
                nv = input('\n%s\nEnter value of %s [default %s]? ' % (desc,s,v))
                if nv: v = nv
            _substitute_values(s,v,known,V)
        if V:
            sys.stderr.write('These values are indeterminate %s\n'%' '.join([x[0] for x in V]))
            os._exit(-1)
        for s in meta_data['substitutables']:
            S = meta_data[s]
            for k,v in list(known.items()):
                r = '{{%s}}'%k
                if type(S) is type([]):
                    for x in range(len(S)):
                        S[x] = _replace(S[x],r,v)
                else:
                    S = _replace(S,r,v)
            meta_data[s] = S
        h40 = '#'*40
        self.showMessage('\n%s\nThe following settings are in effect' % h40)
        for k,v in list(known.items()):
            self.showMessage('%s=%s' % (k,v))

        self._known = known

        self._log('\n%s\nActions' % h40)
        for s in meta_data['actions']:
            self._log(s)

        nv = forceInstall and 'y' or self.yesnoquery('Proceed with install [Y]? ')
        if nv and nv.lower() not in ['y','yes']:
            sys.stderr.write('Install cancelled by user\n')
            os._exit(-1)
        self._log('\n%s\nInstalled Files and Directories' % h40)
        for s in meta_data['actions']:
            try:
                rl_exec(s)
            except:
                self.showMessage('\n!!!!!!!!!!!!!! Error executing action\n"%s"' % s )
                raise
        self.special_test()
        self._log('\n%s\nThese files were retained as unchanged' % h40)
        for s in self._retained:
            self._log('Retained "%s"' % s)
        self.make_version()
        self.remove(*self._unremoved)
        self.removeD(*self._unremovedD)
        self.showMessage('File "%s"\ncontains a log of the installed files and directories!' % self._logfilename,dolog=0)

    def create_dirs(self,dirs=[], perms=0o777):
        for x in dirs:
            try:
                os.makedirs(x,perms)
                self._log('makedirs(%r,0%o)' % (x,perms))
            except:
                self.chmod(perms,x)

    def special_test(self):
        try:
            import pyRXP
            v = float(pyRXP.version)
            if v<=1.02: self.showMessage('\n!!!!!!!!!!!!!! pyRXP.version=%.2f UPGRADE NOW to 1.03 or better!'% v)
        except:
            self.showMessage('\n!!!!!!!!!!!!!! Unable to import pyRXP')

    def chmod(self,mode,*path):
        for p in path:
            try:
                os.chmod(p,mode)
                self._log('chmod(%r,0%o)-->0%o' % (p,mode,os.stat(p).st_mode))
            except:
                self.showMessage('\n!!!!!!!!!!!!!! FAILED chmod 0%o %s'% (mode, p))

    def _unr_extend(self,unr):
        precious = self._precious
        fn = []
        for p in precious:
            fn.extend(fnmatch.filter(unr,p))
        for p in fn:
            unr.remove(p)
        return unr

    def remove(self,*path):
        unr = self._unremoved = self._unr_extend(self._unremoved)
        for p in path:
            for f in glob(p):
                fn = os.path.abspath(os.path.normpath(os.path.normcase(f)))
                if fn in unr:
                    try:
                        kill(fn)
                        self._log('Removed "%s"' % fn)
                        unr.remove(fn)
                    except:
                        self._log('\n!!!!!!!!!!!!!! FAILED remove "%s"'%fn)

    def forced_remove(self,*path):
        for p in path:
            for f in glob(p):
                fn = os.path.abspath(os.path.normpath(os.path.normcase(f)))
                kill(fn)
                if os.path.isfile(fn) or os.path.isdir(fn):
                    self._log('\n!!!!!!!!!!!!!! FAILED forced_remove "%s"'%fn)

    def removeD(self,*path):
        und = self._unremovedD = self._unr_extend(self._unremovedD)
        for p in path:
            for f in glob(p):
                fn = os.path.abspath(os.path.normpath(os.path.normcase(f)))
                if fn in und:
                    if not os.listdir(fn):
                        try:
                            os.rmdir(fn)
                            self._log('Removed "%s"' % fn)
                            und.remove(fn)
                        except:
                            self._log('\n!!!!!!!!!!!!!! FAILED remove "%s"'%fn)

    def make_version(self,appdir=None):
        if not appdir: appdir = self._known['APPDIR']
        project_version = self._meta_data['project_version']
        self.remove(path_join(appdir,'version.txt'))
        data = 'This folder contains %s installed at %4d%02d%02d%02d%02d%02d GMT' % ((project_version,)+tuple(time.gmtime()[:6]))
        self.writefile(path_join(appdir,'version.txt'),'t',data)

    def checkdirs(self,*path):
        for p in path:
            if isDir(p): continue
            if isFile(p):
                self.showMessage('\n!!!!!!!!!!!!!! FAILED checkdirs "%s" is a file!'% p)
            else:
                self.showMessage('\n!!!!!!!!!!!!!! FAILED checkdirs "%s" doesn\'t exist!'% p)

    def print_meta_data(self,full=0):
        from pprint import pprint
        I = list(self._get_meta_data().items())
        I.sort()
        i = 0
        for k,v in I:
            if full or k not in ('actions','substitutables','values', 'predefines', 'filesrelative'):
                if i: print()
                else: i=1
                print(k)
                if k in ('project_version','top_info'):
                    print(v)
                else:
                    pprint(v)
        import time, sys, os
        try:
            from hashlib import md5
        except ImportError:
            from md5 import md5
        from reportlab.pdfbase import pdfutils
        from rlextra.utils.ubold import W
        xtra = []
        for product in ('ReportLabPLUS','RML2PDF', 'PageCatcher', 'PdfEncrypt'):
            try:
                rl_exec("import %sLicense as mod" % product, locals())
                assert os.path.basename(self._fn)==os.path.basename(os.path.dirname(mod.__file__))
            except:
                continue
            try:
                SS = " %s is a licensed product of ReportLab http://www.reportlab.com. " % product
                number = str(mod.number)
                if product=='ReportLabPLUS':
                    if len(number)!=14:
                        xtra.append('%s license number bad!' % product)
                        continue
                    SS = SS.replace('PLUS',' PLUS')
                else:
                    if len(number)<11:
                        xtra.append('%s license number bad!' % product)
                        continue
                    if product=='RML2PDF': SS = SS.replace(' of ',' from ')
                    elif product=='PdfEncrypt': SS = SS[1:]
                user = mod.user
                key = mod.key
                expire = mod.expire
                misc = mod.misc
                Sn = SS+number+SS+user+SS+repr(expire)+SS+repr(misc)
                code = md5(Sn).digest()
                code85 = pdfutils._AsciiBase85Encode(code)
                if code85!=key:
                    xtra.append("%s license Key bad!" % product)
                    continue
                if product!='ReportLabPLUS':
                    if expire is not None and expire<time.time():
                        xtra.append("%s license expired %s" % (product, time.strftime('%Y%m%d',time.gmtime(expire))))
                        continue
                else:
                    if expire is None:
                        if number!=W.AMN or user!=W.AMU:
                            #number is a date
                            import rlextra
                            from datetime import datetime
                            def gd(n):
                                return datetime(int(n[0:4]),int(n[4:6]),int(n[6:8]))
                            try:
                                d = rlextra.Version.split('.')[-1]
                                if len(d)==14: #daily
                                    d = (gd(d) - gd(number)).days
                                else:
                                    d = rlextra.__version__.split()[3].replace('-','')
                                    d = (gd(d) - gd(number)).days
                            except:
                                xtra.append(W.ANLC)
                                continue
                            else:
                                if d>366:
                                    xtra.append(W.ANLM)
                                    continue
                    elif expire<time.time():
                        xtra.append(W.pp('ALE',W.ALE % time.strftime('%Y%m%d',time.gmtime(expire))))
                        continue
                xtra.append("%s license valid!" % product)
            except:
                et, ev= list(map(str,sys.exc_info()[:2]))
                xtra.append("%s license imported, but got error %s(%s)" % (product,et,ev))
        if xtra: print('\n','\n'.join(xtra))
